from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class AgentCommand(Command, command_name="agent"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        agent_handler = session_context.agent_handler

        if agent_handler.agent_enabled:
            agent_handler.disable_agent_mode()
            stream.send("Agent mode disabled", style="success")
        else:
            await agent_handler.enable_agent_mode()
            stream.send("Agent mode enabled", style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Toggle agent mode. In agent mode Mentat will automatically make changes" " and run commands."



from __future__ import annotations

import json
import re
from pathlib import Path

import attr

from mentat.errors import SampleError
from mentat.sampler import __version__


@attr.define
class Sample:
    # TODO: enforce required fields
    title: str = attr.field(default="")
    description: str = attr.field(default="")
    id: str = attr.field(default="")
    parent_id: str = attr.field(default="")
    repo: str = attr.field(default="")
    environment_setup_commit: str | None = attr.field(default=None)
    merge_base: str | None = attr.field(default=None)
    diff_merge_base: str = attr.field(default="")
    diff_active: str = attr.field(default="")
    message_history: list[dict[str, str]] = attr.field(default=[])  # type: ignore
    message_prompt: str = attr.field(default="")
    hint_text: str = attr.field(default="")
    message_edit: str = attr.field(default="")
    context: list[str] = attr.field(default=[])  # type: ignore
    diff_edit: str = attr.field(default="")
    test_patch: str = attr.field(default="")
    test_command: str = attr.field(default="")
    PASS_TO_PASS: str = attr.field(default="")
    version: str = attr.field(default=__version__)

    def save(self, fname: str | Path) -> None:
        with open(Path(fname), "w") as f:
            json.dump(attr.asdict(self), f, indent=4)

    @classmethod
    def load(cls, fname: str | Path) -> Sample:
        with open(fname, "r") as f:
            kwargs = json.load(f)
            _version = kwargs.get("version")
            if _version and _version < "0.2.0":
                kwargs["message_history"] = kwargs.get("message_history", [])[::-1]
                kwargs["version"] = "0.2.0"
                _version = kwargs["version"]
            if _version < "0.3.0":
                # Additional fields from SWE-Bench
                kwargs["environment_setup_commit"] = ""
                kwargs["hint_text"] = ""
                kwargs["test_patch"] = ""
                kwargs["PASS_TO_PASS"] = ""
                kwargs["version"] = "0.3.0"
            if _version != __version__:
                raise SampleError(
                    f"Warning: sample version ({_version}) does not match current" f" version ({__version__})."
                )
            return cls(**kwargs)

    @classmethod
    def from_swe_bench(cls, benchmark: dict[str, str]) -> Sample:
        """Create a Sample from a SWE-Bench benchmark.

        SWE-Bench Fields (https://huggingface.co/datasets/princeton-nlp/SWE-bench#dataset-structure)
        - instance_id: (str) - A formatted instance identifier, usually as repo_owner__repo_name-PR-number.
        - patch: (str) - The gold patch, the patch generated by the PR (minus test-related code), that resolved
          the issue.
        - repo: (str) - The repository owner/name identifier from GitHub.
        - base_commit: (str) - The commit hash of the repository representing the HEAD of the repository before the
          solution PR is applied.
        - hints_text: (str) - Comments made on the issue prior to the creation of the solution PR’s first commit
          creation date.
        - created_at: (str) - The creation date of the pull request.
        - test_patch: (str) - A test-file patch that was contributed by the solution PR.
        - problem_statement: (str) - The issue title and body.
        - version: (str) - Installation version to use for running evaluation.
        - environment_setup_commit: (str) - commit hash to use for environment setup and installation.
        - FAIL_TO_PASS: (str) - A json list of strings that represent the set of tests resolved by the PR and tied to
          the issue resolution.
        - PASS_TO_PASS: (str) - A json list of strings that represent tests that should pass before and after the PR
          application.
        """
        patch = benchmark.get("patch", "")
        edited_files = re.findall(r"diff --git a/(.*?) b/\1", patch)
        return cls(
            title=f"SWE-bench-{benchmark['instance_id']}",
            description="",
            id=benchmark["instance_id"],
            parent_id="",
            repo=f"https://github.com/{benchmark.get('repo')}",
            environment_setup_commit=benchmark.get("environment_setup_commit", ""),
            merge_base=benchmark.get("base_commit"),
            diff_merge_base="",
            diff_active="",
            message_history=[],
            message_prompt=benchmark.get("problem_statement", ""),
            hint_text=benchmark.get("hint_text", ""),
            message_edit="",
            context=edited_files,
            diff_edit=patch,
            test_patch=benchmark.get("test_patch", ""),
            test_command=(
                "" if not benchmark.get("FAIL_TO_PASS") else "pytest " + " ".join(json.loads(benchmark["FAIL_TO_PASS"]))
            ),
            PASS_TO_PASS=(
                "" if not benchmark.get("PASS_TO_PASS") else "pytest " + " ".join(json.loads(benchmark["PASS_TO_PASS"]))
            ),
        )



import json
from pathlib import Path
from typing import List

from typing_extensions import override

from mentat.auto_completer import get_command_filename_completions
from mentat.command.command import Command, CommandArgument
from mentat.errors import PathValidationError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path


class LoadCommand(Command, command_name="load"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        context_file_path = mentat_dir_path / "context.json"

        if len(args) > 1:
            stream.send("Only one context file can be loaded at a time", style="warning")
            return

        if args:
            path_arg = args[0]
            try:
                context_file_path = Path(path_arg).expanduser().resolve()
            except RuntimeError as e:
                raise PathValidationError(f"Invalid context file path provided: {path_arg}: {e}")

        try:
            with open(context_file_path, "r") as file:
                parsed_include_files = json.load(file)
        except FileNotFoundError:
            stream.send(f"Context file not found at {context_file_path}", style="error")
            return
        except json.JSONDecodeError as e:
            stream.send(
                f"Failed to parse context file at {context_file_path}: {e}",
                style="error",
            )
            return

        code_context.from_simple_context_dict(parsed_include_files)

        stream.send(f"Context loaded from {context_file_path}", style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("optional", ["path"])]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return get_command_filename_completions(arguments[-1])

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Loads a context file. Loaded context adds to existing context, it does not replace it."



import json
from enum import Enum
from json import JSONDecodeError
from pathlib import Path
from typing import Any

from typing_extensions import override

from mentat.code_file_manager import CodeFileManager
from mentat.errors import ModelError
from mentat.parsers.change_display_helper import DisplayInformation, FileActionType
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import ParsedLLMResponse, Parser
from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT

block_parser_prompt_filename = Path("block_parser_prompt.txt")


class _BlockParserAction(Enum):
    Insert = "insert"
    Replace = "replace"
    Delete = "delete"
    CreateFile = "create-file"
    DeleteFile = "delete-file"
    RenameFile = "rename-file"


class _BlockParserIndicator(Enum):
    Start = "@@start"
    Code = "@@code"
    End = "@@end"


class _BlockParserJsonKeys(Enum):
    File = "file"
    Action = "action"
    Name = "name"
    BeforeLine = "insert-before-line"
    AfterLine = "insert-after-line"
    StartLine = "start-line"
    EndLine = "end-line"


class _BlockParserDeserializedJson:
    def __init__(self, json_data: dict[str, Any]):
        self.file = json_data.get(_BlockParserJsonKeys.File.value, None)
        self.action = json_data.get(_BlockParserJsonKeys.Action.value, None)
        self.name = json_data.get(_BlockParserJsonKeys.Name.value, None)
        self.before_line = json_data.get(_BlockParserJsonKeys.BeforeLine.value, None)
        self.after_line = json_data.get(_BlockParserJsonKeys.AfterLine.value, None)
        self.start_line = json_data.get(_BlockParserJsonKeys.StartLine.value, None)
        self.end_line = json_data.get(_BlockParserJsonKeys.EndLine.value, None)

        if self.file is not None:
            self.file = Path(self.file)
        if self.action is not None:
            self.action: _BlockParserAction | None = _BlockParserAction(self.action)
        if self.name is not None:
            self.name = Path(self.name)
        if self.before_line is not None:
            self.before_line = int(self.before_line)
        if self.after_line is not None:
            self.after_line = int(self.after_line)
        if self.start_line is not None:
            self.start_line = int(self.start_line)
        if self.end_line is not None:
            self.end_line = int(self.end_line)


class BlockParser(Parser):
    @override
    def get_system_prompt(self) -> str:
        return read_prompt(block_parser_prompt_filename)

    @override
    def _could_be_special(self, cur_line: str) -> bool:
        return any(to_match.value.startswith(cur_line.strip()) for to_match in _BlockParserIndicator) and (
            bool(cur_line.strip()) or not cur_line.endswith("\n")
        )

    @override
    def _starts_special(self, line: str) -> bool:
        return line == _BlockParserIndicator.Start.value

    @override
    def _ends_special(self, line: str) -> bool:
        return line == _BlockParserIndicator.Code.value or line == _BlockParserIndicator.End.value

    @override
    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        cwd: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        block = special_block.strip().split("\n")
        json_lines = block[1:-1]
        try:
            json_data: dict[str, Any] = json.loads("\n".join(json_lines))
            deserialized_json = _BlockParserDeserializedJson(json_data)
        except (JSONDecodeError, ValueError):
            raise ModelError("Error: Model output malformed json.")

        if deserialized_json.action is None:
            raise ModelError("Error: Model output malformed json.")

        starting_line = 0
        ending_line = 0
        match deserialized_json.action:
            case _BlockParserAction.Insert:
                if deserialized_json.before_line is not None:
                    starting_line = deserialized_json.before_line - 1
                    if deserialized_json.after_line is not None and starting_line != deserialized_json.after_line:
                        raise ModelError("Error: Model output malformed edit.")
                elif deserialized_json.after_line is not None:
                    starting_line = deserialized_json.after_line
                else:
                    raise ModelError("Error: Model output malformed edit.")
                ending_line = starting_line
                file_action = FileActionType.UpdateFile

            case _BlockParserAction.Replace | _BlockParserAction.Delete:
                if deserialized_json.start_line is None or deserialized_json.end_line is None:
                    raise ModelError("Error: Model output malformed edit.")
                starting_line = deserialized_json.start_line - 1
                ending_line = deserialized_json.end_line
                file_action = FileActionType.UpdateFile

            case _BlockParserAction.CreateFile:
                file_action = FileActionType.CreateFile

            case _BlockParserAction.DeleteFile:
                file_action = FileActionType.DeleteFile

            case _BlockParserAction.RenameFile:
                file_action = FileActionType.RenameFile
        if ending_line < starting_line:
            raise ModelError("Error: Model output malformed edit.")

        full_path = (cwd / deserialized_json.file).resolve()
        rename_file_path = (cwd / deserialized_json.name).resolve() if deserialized_json.name else None

        file_lines = self._get_file_lines(code_file_manager, rename_map, full_path)
        display_information = DisplayInformation(
            deserialized_json.file,
            file_lines,
            [],
            file_lines[starting_line:ending_line],
            file_action,
            starting_line,
            ending_line,
            deserialized_json.name,
        )

        replacements = list[Replacement]()
        if deserialized_json.action == _BlockParserAction.Delete:
            replacements.append(Replacement(starting_line, ending_line, []))
        file_edit = FileEdit(
            full_path,
            replacements,
            is_creation=file_action == FileActionType.CreateFile,
            is_deletion=file_action == FileActionType.DeleteFile,
            rename_file_path=rename_file_path,
        )
        has_code = block[-1] == _BlockParserIndicator.Code.value
        return (display_information, file_edit, has_code)

    @override
    def _ends_code(self, line: str) -> bool:
        return line == _BlockParserIndicator.End.value

    @override
    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ):
        file_edit.replacements.append(
            Replacement(
                display_information.first_changed_line,
                display_information.last_changed_line,
                code_block.split("\n")[:-2],
            )
        )

    def file_edits_to_llm_message(self, parsedLLMResponse: ParsedLLMResponse) -> str:
        """
        Inverse of stream_and_parse_llm_response
        """
        session_context = SESSION_CONTEXT.get()

        ans = parsedLLMResponse.conversation.strip() + "\n\n"
        for file_edit in parsedLLMResponse.file_edits:
            tmp = {}
            tmp[_BlockParserJsonKeys.File.value] = file_edit.file_path.relative_to(session_context.cwd).as_posix()
            if file_edit.is_creation:
                tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.CreateFile.value
            elif file_edit.is_deletion:
                tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.DeleteFile.value
            elif file_edit.rename_file_path is not None:
                tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.RenameFile.value
                tmp[_BlockParserJsonKeys.Name.value] = file_edit.rename_file_path.relative_to(
                    session_context.cwd
                ).as_posix()
            if _BlockParserJsonKeys.Action.value in tmp:
                ans += _BlockParserIndicator.Start.value + "\n"
                ans += json.dumps(tmp, indent=4) + "\n"
                if tmp[_BlockParserJsonKeys.Action.value] == _BlockParserAction.CreateFile.value:
                    ans += _BlockParserIndicator.Code.value + "\n"
                    ans += "\n".join(file_edit.replacements[0].new_lines) + "\n"
                ans += _BlockParserIndicator.End.value + "\n"
            if not file_edit.is_creation:
                for replacement in file_edit.replacements:
                    tmp = {}
                    tmp[_BlockParserJsonKeys.File.value] = file_edit.file_path.relative_to(
                        session_context.cwd
                    ).as_posix()
                    ans += _BlockParserIndicator.Start.value + "\n"
                    starting_line = replacement.starting_line
                    ending_line = replacement.ending_line
                    if len(replacement.new_lines) == 0:
                        tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.Delete.value
                        tmp[_BlockParserJsonKeys.StartLine.value] = starting_line + 1
                        tmp[_BlockParserJsonKeys.EndLine.value] = ending_line
                    else:
                        if starting_line == ending_line:
                            tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.Insert.value
                            tmp[_BlockParserJsonKeys.AfterLine.value] = starting_line
                            tmp[_BlockParserJsonKeys.BeforeLine.value] = ending_line + 1
                        else:
                            tmp[_BlockParserJsonKeys.Action.value] = _BlockParserAction.Replace.value
                            tmp[_BlockParserJsonKeys.StartLine.value] = starting_line + 1
                            tmp[_BlockParserJsonKeys.EndLine.value] = ending_line
                    ans += json.dumps(tmp, indent=4) + "\n"
                    if len(replacement.new_lines) > 0:
                        ans += _BlockParserIndicator.Code.value + "\n"
                        ans += "\n".join(replacement.new_lines) + "\n"
                    ans += _BlockParserIndicator.End.value + "\n"

        return ans



import os
import platform
from uuid import UUID, uuid4

import sentry_sdk

from mentat.app_conf import IS_DEV
from mentat.utils import mentat_dir_path
from mentat.version import __version__

user_id_path = mentat_dir_path / ".user"


def _new_user() -> UUID:
    user = uuid4()
    with user_id_path.open("w") as uuid_file:
        uuid_file.write(str(user))
    return user


def _get_user() -> UUID:
    if user_id_path.exists():
        with user_id_path.open("r") as uuid_file:
            try:
                user = UUID(uuid_file.read())
            except ValueError:
                user = _new_user()
    else:
        user = _new_user()
    return user


def sentry_init():
    # Check if in dev or testing (although we should never be in testing in prod)
    if IS_DEV or "PYTEST_CURRENT_TEST" in os.environ:
        return

    sentry_sdk.init(
        dsn="https://fa4d2b80dab0938c38c89384dc317f1b@o4506146491006976.ingest.sentry.io/4506187066245120",
        profiles_sample_rate=1.0,
        enable_tracing=True,
        traces_sample_rate=1.0,
    )
    sentry_sdk.set_tag("version", __version__)
    sentry_sdk.set_user({"id": _get_user()})
    uname = platform.uname()
    sentry_sdk.set_context(
        "os",
        {
            "raw_description": " ".join(uname),
            "name": uname.system,
            "version": uname.release,
        },
    )



themes: dict[str, dict[str, str]] = {
    "light": {
        "prompt": "gray",
        "code": "blue",
        "info": "cyan",
        "failure": "dark_red",
        "success": "green",
        "input": "bright_blue",
        "error": "red",
        "warning": "yellow",
    },
    "dark": {
        "prompt": "white",
        "code": "blue",
        "info": "cyan",
        "failure": "bright_red",
        "success": "green",
        "input": "bright_blue",
        "error": "red",
        "warning": "yellow",
    },
}



from typing import Iterable

from mentat.code_feature import CodeFeature
from mentat.feature_filters.feature_filter import FeatureFilter


class TruncateFilter(FeatureFilter):
    def __init__(
        self,
        max_tokens: int,
        model: str = "gpt-4",
    ):
        self.max_tokens = max_tokens
        self.model = model

    async def filter(
        self,
        features: Iterable[CodeFeature],
    ) -> list[CodeFeature]:
        """Truncate the features to max_token tokens."""
        output = list[CodeFeature]()
        remaining_tokens = self.max_tokens
        for feature in features:
            if feature.count_tokens(self.model) <= remaining_tokens:
                output.append(feature)
                remaining_tokens -= feature.count_tokens(self.model)

        return output



from dataclasses import dataclass
from typing import List

from textual import events
from textual_autocomplete import AutoComplete, Dropdown, DropdownItem
from typing_extensions import override

from mentat.auto_completer import Completion
from mentat.session_stream import SessionStream, StreamMessageSource
from mentat.terminal.history_suggester import HistorySuggester


class PatchedAutoComplete(AutoComplete):
    """
    We want more fine control over replacing the text when a completion is
    selected, so we patch _select_item.
    """

    @override
    def _select_item(self):
        selected: CompletionDropdownItem | None = self.dropdown.selected_item  # pyright: ignore
        if self.dropdown.display and selected is not None:
            cursor_position: int = self.input.cursor_position  # pyright: ignore
            before = self.input.value[: cursor_position + selected.position]
            after = self.input.value[cursor_position:]
            self.input.value = before + selected.content + after
            self.input.cursor_position = cursor_position + selected.position + len(selected.content)

            self.dropdown.display = False
            self.post_message(self.Selected(item=selected))


@dataclass
class CompletionDropdownItem(DropdownItem):
    """
    A dropdown item with a full completion to carry all of the information we need
    """

    content: str = ""
    position: int = 0


class PatchedDropdown(Dropdown):
    """
    The dropdown class in textual_autocomplete isn't able to do everything we need it to (mainly async completions),
    so this class patchs some of its methods to add the functionality we need.
    """

    @override
    def __init__(
        self,
        stream: SessionStream,
        history_suggestor: HistorySuggester,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.stream = stream
        self.history_suggester = history_suggestor
        super().__init__([], id, classes)

    # This function is mostly copied from Dropdown; the only change is using our async functions as callbacks
    @override
    def on_mount(self, event: events.Mount) -> None:
        screen_layers = list(self.screen.styles.layers)
        if "textual-autocomplete" not in screen_layers:
            screen_layers.append("textual-autocomplete")

        self.screen.styles.layers = tuple(screen_layers)  # type: ignore

        self.watch(
            self.input_widget,
            attribute_name="value",
            callback=self._async_input_value_changed,  # Changed
        )

        self.watch(
            self.input_widget,
            attribute_name="cursor_position",
            callback=self._async_input_cursor_position_changed,  # Changed
        )

        self.watch(
            self.screen,
            attribute_name="scroll_target_y",
            callback=self.handle_screen_scroll,
        )

        if self.input_widget is not None:  # pyright: ignore
            self.sync_state(
                self.input_widget.value,
                self.input_widget.cursor_position,  # pyright: ignore
            )

    async def _get_completions(self) -> List[DropdownItem]:
        text = self.input_widget.value[
            : self.input_widget.cursor_position  # pyright: ignore
        ]
        message = self.stream.send(
            text,
            source=StreamMessageSource.CLIENT,
            channel="completion_request",
            command_autocomplete=self.parent.parent.command_autocomplete,  # pyright: ignore
        )

        response = await self.stream.recv(channel=f"completion_request:{message.id}")
        completions: List[Completion] = response.data
        return [
            CompletionDropdownItem(
                content=completion["content"],
                position=completion["position"],
                main=(completion["display"] if completion["display"] is not None else completion["content"]),
            )
            for completion in completions
        ]

    # These 2 functions are very similar to their synchronous versions
    async def _async_input_cursor_position_changed(self, cursor_position: int) -> None:
        if (
            self.input_widget is not None  # pyright: ignore
            and not self.history_suggester.just_moved(self.input_widget.value)
        ):
            matches = await self._get_completions()
            self._mentat_sync_state(self.input_widget.value, cursor_position, matches)

    async def _async_input_value_changed(self, value: str) -> None:
        if (
            self.input_widget is not None  # pyright: ignore
            and not self.history_suggester.just_moved(value)
        ):
            matches = await self._get_completions()
            self._mentat_sync_state(
                value,
                self.input_widget.cursor_position,
                matches,  # pyright: ignore
            )

    # Since we do the autocomplete matching on the backend, we completely change this function
    # to not do any matching of its own
    def _mentat_sync_state(self, value: str, input_cursor_position: int, matches: List[DropdownItem]):
        self.child.matches = matches
        if matches:
            self.styles.width = max(len(match.main) for match in matches) + 2

        self.display = len(matches) > 0 and value != "" and self.input_widget.has_focus
        self.cursor_home()
        self.reposition(input_cursor_position)
        self.child.refresh()

    # By default the dropdown can only be below the input, but since our input is always at the bottom of the screen
    # we want the opposite, so we patch this function to drop up instead of down
    @override
    def reposition(
        self,
        input_cursor_position: int | None = None,
        scroll_target_adjust_y: int = 0,
    ) -> None:
        if self.input_widget is None:  # pyright: ignore
            return

        if input_cursor_position is None:
            input_cursor_position = self.input_widget.cursor_position  # pyright: ignore

        top, right, bottom, left = self.styles.margin  # pyright: ignore
        x, y, width, height = self.input_widget.content_region  # pyright: ignore
        # This is the line we change to fix the position
        line_below_cursor = (
            y
            + scroll_target_adjust_y
            - min(
                int(self.styles.max_height.value),  # pyright: ignore
                len(self.child.matches),
            )
        )

        cursor_screen_position = x + (  # pyright: ignore
            input_cursor_position - self.input_widget.view_position
        )
        self.styles.margin = (
            line_below_cursor,
            right,
            bottom,
            cursor_screen_position,
        )



from configparser import ConfigParser
from pathlib import Path

from mentat.utils import fetch_resource

conf_ini_path = Path("conf/conf.ini")

config = ConfigParser()
with fetch_resource(conf_ini_path).open("r") as conf_ini_file:
    config.read_file(conf_ini_file)

ENV_TYPE = config["environment"]["type"]
IS_DEV = ENV_TYPE == "dev"
IS_PROD = ENV_TYPE == "prod"



from __future__ import annotations

import os
from asyncio import Event
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from rich.console import RenderableType
from rich.markup import escape
from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Input, ProgressBar, RichLog, Static, Tree
from textual.widgets._tree import TreeNode
from typing_extensions import override

from mentat.session_stream import SessionStream, StreamMessage
from mentat.terminal.history_suggester import HistorySuggester
from mentat.terminal.patched_autocomplete import PatchedAutoComplete, PatchedDropdown
from mentat.terminal.themes import themes
from mentat.utils import fetch_resource, mentat_dir_path

if TYPE_CHECKING:
    from mentat.terminal.client import TerminalClient

css_path = Path("textual/terminal_app.tcss")
history_file_location = mentat_dir_path / "prompt_history"

change_delimiter = "=" * 60


class ContentContainer(Static):
    BINDINGS = [
        ("up", "history_up", "Move up in history"),
        ("down", "history_down", "Move down in history"),
    ]

    def __init__(
        self,
        stream: SessionStream,
        theme: dict[str, str],
        renderable: RenderableType = "",
        **kwargs: Any,
    ):
        self.stream = stream
        self.theme = theme
        self.input_event = Event()
        self.last_user_input = ""
        self.suggester = HistorySuggester(history_file=history_file_location)
        self.loading_bar = None
        self.cur_line = ""

        super().__init__(renderable, **kwargs)

    @override
    def compose(self) -> ComposeResult:
        self.content = RichLog(wrap=True, markup=True, auto_scroll=True)
        yield self.content
        # RichLogs can't edit existing lines, only add new ones, so we have a 'buffer' widget until we reach a new line.
        self.last_content = Static(self.cur_line, classes="content-piece")
        yield self.last_content
        yield PatchedAutoComplete(
            Input(
                classes="user-input",
                disabled=True,
                suggester=self.suggester,
            ),
            PatchedDropdown(self.stream, self.suggester),
        )
        self.add_content("Type 'q' or use Ctrl-C to quit at any time.\n", color="cyan")
        self.add_content("What can I do for you?\n", color="blue")

    @on(Input.Submitted)
    def on_user_input(self, event: Input.Submitted):
        self.last_user_input = event.value
        self.suggester.append_to_history(event.value)
        self.input_event.set()

    async def collect_user_input(self, default_prompt: str) -> str:
        self.input_event.clear()

        user_input = self.query_one(Input)
        user_input.disabled = False
        user_input.value = default_prompt
        user_input.focus()
        await self.input_event.wait()
        user_input.value = ""
        user_input.disabled = True

        self.add_content(f">>> {self.last_user_input}\n", color=self.theme["prompt"])
        return self.last_user_input

    def action_history_up(self):
        history = self.suggester.move_up()
        if history is not None:
            user_input = self.query_one(Input)
            user_input.value = history
            user_input.cursor_position = len(history)

    def action_history_down(self):
        history = self.suggester.move_down()
        if history is not None:
            user_input = self.query_one(Input)
            user_input.value = history
            user_input.cursor_position = len(history)

    def start_loading(self):
        if not self.loading_bar:
            self.loading_bar = ProgressBar(id="loading-display", show_percentage=False, show_eta=False)
            self.mount(self.loading_bar)

    def end_loading(self):
        if self.loading_bar is not None:
            self.loading_bar.remove()
            self.loading_bar = None

    def add_content(self, new_content: str, color: str | None = None):
        new_content = escape(new_content)
        lines = [f"[{color}]{line}[/{color}]" if color else line for line in new_content.split("\n")]
        for line in lines[:-1]:
            line = self.cur_line + line
            self.cur_line = ""
            self.content.write(line)
        self.cur_line += lines[-1]
        self.last_content.update(self.cur_line)
        self.scroll_end(animate=False)


class ContextContainer(Static):
    def _build_path_tree(self, files: list[str], cwd: Path):
        """Builds a tree of paths from a list of CodeFiles."""
        tree = dict[str, Any]()
        for file in files:
            path = os.path.relpath(file, cwd)
            parts = Path(path).parts
            current_level = tree
            for part in parts:
                if part not in current_level:
                    current_level[part] = {}
                current_level = current_level[part]
        return tree

    def _build_sub_tree(
        self,
        cur_path: Path,
        root: TreeNode[Any],
        children: Dict[str, Any],
        git_diff_paths: Set[Path],
        untracked_paths: Set[Path],
        untracked: bool = False,
    ):
        for child, grandchildren in children.items():
            new_path = cur_path / child
            path_untracked = new_path in untracked_paths or untracked
            if path_untracked:
                label = f"[red]! {child}[/red]"
            else:
                label = child
            if not grandchildren:
                if new_path in git_diff_paths:
                    label = f"[green]* {child}[/green]"
                root.add_leaf(label)
            else:
                child_node = root.add(label, expand=True)
                self._build_sub_tree(
                    new_path,
                    child_node,
                    grandchildren,
                    git_diff_paths,
                    untracked_paths,
                    path_untracked,
                )

    def _build_tree_widget(
        self,
        files: list[str],
        cwd: Path,
        git_diff_paths: Set[Path],
        untracked_paths: Set[Path],
    ) -> Tree[Any]:
        path_tree = self._build_path_tree(files, cwd)
        tree: Tree[Any] = Tree(f"[blue]{cwd.name}[/blue]")
        tree.root.expand()
        self._build_sub_tree(cwd, tree.root, path_tree, git_diff_paths, untracked_paths)
        return tree

    def update_context(
        self,
        cwd: Path,
        diff_context_display: Optional[str],
        auto_context_tokens: int,
        features: List[str],
        git_diff_paths: Set[Path],
        git_untracked_paths: Set[Path],
        total_tokens: int,
        total_cost: float,
    ):
        feature_tree = self._build_tree_widget(features, cwd, git_diff_paths, git_untracked_paths)

        context_header = ""
        context_header += "[blue bold]Code Context:[/blue bold]"
        context_header += f"\nTokens: [yellow]{total_tokens}[/yellow]"
        context_header += f"\nTotal Session Cost: [yellow]${total_cost:.2f}[/yellow]"
        context_header += f"\nDirectory: {cwd}"
        if diff_context_display:
            context_header += f"\nDiff:[green]{diff_context_display}[/green]"
        if auto_context_tokens > 0:
            context_header += f"\nAuto-Context: Enabled\nAuto-Context Tokens: {auto_context_tokens}"
        else:
            context_header += "\nAuto-Context: [yellow]Disabled[/yellow]"

        context_header += "\nIncluded Files:"
        if not features:
            context_header += " [yellow]None[/yellow]"

        self.remove_children()
        self.mount(Static(context_header))
        if features:
            self.mount(feature_tree)


css_resource = fetch_resource(css_path)
with css_resource.open("r") as css_file:
    css = css_file.read()


class TerminalApp(App[None]):
    BINDINGS = [("ctrl+c", "on_interrupt", "Send interrupt")]
    CSS = css
    TITLE = "Mentat"

    def __init__(self, client: TerminalClient, **kwargs: Any):
        self.client = client
        self.command_autocomplete = False
        self.last_filepath = None
        super().__init__(**kwargs)

    @override
    def compose(self) -> ComposeResult:
        self.dark = self.client.config.theme == "dark"
        self.theme = themes[self.client.config.theme]
        self.content_container = ContentContainer(self.client.session.stream, self.theme)
        yield self.content_container
        yield ContextContainer()

    def display_stream_message(self, message: StreamMessage):
        end = "\n"
        color = None
        content_container = self.query_one(ContentContainer)

        if isinstance(message.extra.get("end"), str):
            end = message.extra["end"]
        if isinstance(message.extra.get("color"), str):
            color = message.extra["color"]
        if isinstance(message.extra.get("style"), str):
            style = message.extra["style"]
            color = self.theme[style]
        if message.extra.get("delimiter", False):
            content_container.add_content(f"{change_delimiter}\n")
        filepath = message.extra.get("filepath")
        if filepath != self.last_filepath:
            if self.last_filepath:
                content_container.add_content(f"{change_delimiter}\n\n")
            if filepath:
                filepath_display, filepath_display_type = message.extra.get("filepath_display", filepath)
                content_container.add_content(
                    f"{filepath_display}\n",
                    color=(
                        "bright_green"
                        if filepath_display_type == "creation"
                        else (
                            "bright_red"
                            if filepath_display_type == "deletion"
                            else ("yellow" if filepath_display_type == "rename" else "bright_blue")
                        )
                    ),
                )
                content_container.add_content(f"{change_delimiter}\n")
            self.last_filepath = filepath

        content = str(message.data) + end
        content_container.add_content(content, color)

    async def get_user_input(self, default_prompt: str, command_autocomplete: bool) -> str:
        # This is a really janky way to pass command_autocomplete to Dropdown._get_completions(),
        # but I couldn't think of anything better
        self.command_autocomplete = command_autocomplete
        content_container = self.query_one(ContentContainer)
        return await content_container.collect_user_input(default_prompt)

    def update_context(
        self,
        cwd: Path,
        diff_context_display: Optional[str],
        auto_context_tokens: int,
        features: List[str],
        git_diff_paths: Set[Path],
        git_untracked_paths: Set[Path],
        total_tokens: int,
        total_cost: float,
    ):
        context_container = self.query_one(ContextContainer)
        context_container.update_context(
            cwd,
            diff_context_display,
            auto_context_tokens,
            features,
            git_diff_paths,
            git_untracked_paths,
            total_tokens,
            total_cost,
        )

    def action_on_interrupt(self):
        self.client.send_interrupt()

    def disable_app(self):
        self.query_one(Input).disabled = True
        self.end_loading()

    def start_loading(self):
        self.query_one(ContentContainer).start_loading()

    def end_loading(self):
        self.query_one(ContentContainer).end_loading()



import asyncio
from collections import deque
from typing import Any, Dict, List, Literal, Tuple

from mentat.session_context import SESSION_CONTEXT

# TODO: Make this a class
FormattedString = str | Tuple[str, Dict[str, Any]] | List[Tuple[str, Dict[str, Any]]]


def send_formatted_string(string: FormattedString):
    ctx = SESSION_CONTEXT.get()

    if isinstance(string, List):
        for line in string:
            ctx.stream.send(line[0], **line[1], end="")
        ctx.stream.send("")
    elif isinstance(string, str):
        ctx.stream.send(string)
    else:
        ctx.stream.send(string[0], **string[1])


class StreamingPrinter:
    def __init__(self):
        self.strings_to_print: deque[Tuple[str, Dict[str, Any]]] = deque()
        self.finishing = False
        self.shutdown = False
        self.cur_file: str | None = None
        self.cur_file_display: Tuple[str, Literal["edit", "creation", "deletion", "rename"]] | None = None

    def add_string(
        self,
        formatted_string: FormattedString,
        end: str = "\n",
        allow_empty: bool = False,
    ):
        if self.finishing:
            return

        if isinstance(formatted_string, List):
            for string in formatted_string:
                self.add_string(string, end="")
            if end and (allow_empty or any(string[0] for string in formatted_string)):
                self.add_string(end, end="")
            return
        if isinstance(formatted_string, str):
            string = formatted_string
            styles = {}
        else:
            string = formatted_string[0]
            styles = formatted_string[1]

        styles["filepath"] = self.cur_file
        styles["filepath_display"] = list(self.cur_file_display) if self.cur_file_display else None
        if not string:
            if allow_empty:
                self.strings_to_print.append(("", styles))
            return

        self.strings_to_print.extend((char, styles) for char in string)
        if end:
            self.add_string(end, end="")

    def add_delimiter(self):
        self.add_string(("", {"delimiter": True}), end="", allow_empty=True)

    def sleep_time(self) -> float:
        max_finish_time = 1.0
        required_sleep_time = max_finish_time / (len(self.strings_to_print) + 1)
        max_sleep = 0.002 if self.finishing else 0.006
        min_sleep = 0.001
        return max(min(max_sleep, required_sleep_time), min_sleep)

    async def print_lines(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream

        while not self.shutdown:
            if self.strings_to_print:
                next_string = self.strings_to_print.popleft()
                stream.send(next_string[0], end="", **next_string[1])
            elif self.finishing:
                break
            await asyncio.sleep(self.sleep_time())

    def wrap_it_up(self):
        self.finishing = True

    def shutdown_printer(self):
        self.shutdown = True



import json
import platform
import subprocess
from functools import cache
from pathlib import Path

from mentat.errors import MentatError


@cache
def ensure_ctags_installed() -> None:
    try:
        subprocess.run(
            ["ctags", "--help"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return
    except subprocess.CalledProcessError:
        pass

    os_name = platform.system()
    match os_name:
        case "Linux":
            suggested_install_command = "sudo apt install universal-ctags"
        case "Darwin":  # macOS
            suggested_install_command = "brew install universal-ctags"
        case "Windows":
            suggested_install_command = "choco install universal-ctags"
        case _:
            suggested_install_command = None

    error_message = "Missing Dependency: universal-ctags (required for auto-context)"
    if suggested_install_command:
        error_message += f"\nSuggested install method for your OS: `{suggested_install_command}`"
    error_message += "\nSee README.md for full installation details."
    raise MentatError(error_message)


def get_ctag_lines_and_names(path: Path) -> list[tuple[int, str]]:
    ensure_ctags_installed()

    json_tags = (
        subprocess.check_output(
            ["ctags", "--output-format=json", "--fields=+n", str(path)],
            stderr=subprocess.DEVNULL,
            start_new_session=True,
            text=True,
        )
        .strip()
        .splitlines()
    )

    lines_and_names: list[tuple[int, str]] = []
    for json_tag in json_tags:
        tag_dict = json.loads(json_tag)
        name = tag_dict["name"]
        if "scope" in tag_dict:
            name = f"{tag_dict['scope']}.{name}"
        line = tag_dict["line"]
        lines_and_names.append((line, name))

    return lines_and_names



from pathlib import Path
from textwrap import dedent
from typing import Any, AsyncGenerator, List

from mentat.llm_api_handler import chunk_to_lines
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import ParsedLLMResponse
from mentat.session_context import SESSION_CONTEXT

"""
This parser is primarily intended as a utility for working with samples, rather than as a parser
for the LLM. We implement only those methods/properties which are necessary for that.

Things to be aware of:
1. git diffs include SHA-1 hashes, which link the diff to a known record in the git repo. When
    we generate diffs from scratch, we won't have these, so the diff can't be reverse-applied via
    `git apply myfile.diff`. The SHA-1 hashes should also be removed when comparing outputs in 
    tests as they are not expected to match.

2. git diffs include removed and to-be-modified lines, prefixed with a '-'. These are drawn from
    FileEdit.previous_file_lines field. When calling GitParser.file_edits_to_llm_message, 
    this field must be set (except for new files).

3. Each 'replacement' includes a 'hunk header' line, which is explained below for quick reference:

@@ -a,b +c,d @@

a = original starting line number
b = number of original lines changed, omitted if = 1
c = edited starting line number
d = number of edited lines in the change, ommitted if = 1
"""


class GitParser:
    # This doesn't actually "stream and parse" but it is named this way to match the interface of
    # the production parsers for use in the translation script.
    async def stream_and_parse_llm_response(
        self,
        response: AsyncGenerator[Any, None],
    ) -> ParsedLLMResponse:
        string = ""
        async for chunk in response:
            for content in chunk_to_lines(chunk):
                string += content
        return self.parse_llm_response(string)

    def parse_llm_response(self, content: str) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()

        split_on_diff = content.split("diff --git")
        conversation = split_on_diff[0].strip()
        split_on_diff = split_on_diff[1:]
        git_diff = [f"diff --git{diff}" for diff in split_on_diff]

        # If there's a commit record, remove everything except the commit message.
        if "commit " in conversation and "\n\n" in conversation:
            conversation, commit_record = conversation.split("commit ", 1)
            commit_message = dedent(commit_record.split("\n\n")[1].strip())
            conversation += commit_message

        file_edits: List[FileEdit] = []
        for diff in split_on_diff:
            is_creation = "new file mode" in diff
            is_deletion = "deleted file mode" in diff
            first_line = diff.split("\n")[0]
            start_file_name = Path(first_line.split()[0][2:]).resolve()
            end_file_name = Path(first_line.split()[1][2:]).resolve()
            if start_file_name != end_file_name:
                new_name = end_file_name
            else:
                new_name = None

            file_edit = FileEdit(
                (session_context.cwd / start_file_name).resolve(),
                [],
                is_creation=is_creation,
                is_deletion=is_deletion,
                rename_file_path=new_name,
            )
            if not is_creation:
                file_edit.previous_file_lines = session_context.code_file_manager.file_lines.get(start_file_name, [])
            diff_split = diff.split("\n@@")
            if not is_deletion:
                for change in diff_split[1:]:
                    line_info = change.split("@@")[0]
                    # Git diff line represents line number information:
                    # @@ -a,b +c,d @@
                    # a is the original starting line number and b is the original number of lines.
                    # c and d are the new values.
                    # Both b and d are omitted when 1.
                    a_b = line_info.split()[0].split(",")
                    if is_creation:
                        start_line, end_line = 0, 0
                    else:
                        a = int(a_b[0][1:])
                        b = 1 if len(a_b) == 1 else int(a_b[1])
                        start_line = a - (1 if b > 0 else 0)
                        end_line = start_line + b
                    line_changes = change.split("@@")[1]
                    code_lines = line_changes.split("\n")[1:]  # Discard optional context
                    # This check is necessary because new code sometimes starts on the same line
                    # as @@ sometimes on the next line.
                    if code_lines[0] == "":
                        code_lines = code_lines[1:]
                    if code_lines[-1] == "":
                        code_lines = code_lines[:-1]

                    # Git diff gives context for human readability we don't want to train the llm
                    # to produce.
                    starting_repetition = 0
                    for line in code_lines:
                        if line.startswith(" "):
                            starting_repetition += 1
                        else:
                            break
                    ending_repetition = 0
                    for line in reversed(code_lines):
                        if line.startswith(" "):
                            ending_repetition += 1
                        else:
                            break
                    start_line += starting_repetition
                    end_line -= ending_repetition

                    lines: List[str] = []
                    for line in code_lines[starting_repetition : len(code_lines) - ending_repetition]:
                        if not line.startswith("-"):
                            lines.append(line[1:])

                    file_edit.replacements.append(Replacement(start_line, end_line, lines))

            file_edits.append(file_edit)

        return ParsedLLMResponse(f"{conversation}\n\n{git_diff}", conversation, file_edits)

    def file_edit_to_git_diff(self, file_edit: FileEdit) -> str:
        """Converts a FileEdit object into a git diff string."""
        session_context = SESSION_CONTEXT.get()
        cwd = session_context.cwd

        diff_lines: list[str] = []
        file_path_str = Path(file_edit.file_path).relative_to(session_context.cwd).as_posix()

        if file_edit.is_deletion:
            assert file_edit.previous_file_lines is not None, "Missing previous lines"
            diff_lines.append(f"diff --git a/{file_path_str} b/{file_path_str}")
            diff_lines.append("deleted file mode 100644")
            diff_lines.append("index fffffff..0000000")
            diff_lines.append(f"--- a/{file_path_str}")
            diff_lines.append("+++ /dev/null")
            diff_lines.append(f"@@ -1,{len(file_edit.previous_file_lines) - 1} +0,0 @@")
            for line in file_edit.previous_file_lines:
                diff_lines.append(f"-{line}")
            if diff_lines[-1] == "-":
                diff_lines = diff_lines[:-1]
            return "\n".join(diff_lines)

        if file_edit.is_creation:
            diff_lines.append(f"diff --git a/{file_path_str} b/{file_path_str}")
            diff_lines.append("new file mode 100644")
            diff_lines.append("index 0000000..fffffff")
            diff_lines.append("--- /dev/null")
            diff_lines.append(f"+++ b/{file_path_str}")
        elif file_edit.rename_file_path:
            new_file_path_str = file_edit.rename_file_path.relative_to(cwd).as_posix()
            diff_lines.append(f"diff --git a/{file_path_str} b/{new_file_path_str}")
            diff_lines.append("similarity index 100%")
            diff_lines.append(f"rename from {file_path_str}")
            diff_lines.append(f"rename to {new_file_path_str}")
        else:
            diff_lines.append(f"diff --git a/{file_path_str} b/{file_path_str}")
            diff_lines.append("index fffffff..fffffff 100644")
            diff_lines.append(f"--- a/{file_path_str}")
            diff_lines.append(f"+++ b/{file_path_str}")

        sorted_replacements = sorted(file_edit.replacements, key=lambda r: r.starting_line)
        net_change_in_lines: int = 0
        for replacement in sorted_replacements:
            if file_edit.is_creation:
                a = 0
                b = 0
            else:
                n_changed_lines = replacement.ending_line - replacement.starting_line
                a = replacement.starting_line + (1 if n_changed_lines > 0 else 0)
                b = n_changed_lines
            n_inserted_lines = len(replacement.new_lines)
            c = (
                replacement.starting_line
                + 1  # Git uses one-based indiexing
                - (1 if n_inserted_lines == 0 else 0)
                + net_change_in_lines
            )
            d = n_inserted_lines
            hunk_header = f"@@ -{a}" + (" " if b == 1 else f",{b} ") + f"+{c}" + (" @@" if d == 1 else f",{d} @@")
            net_change_in_lines += d - b
            diff_lines.append(hunk_header)

            if not file_edit.is_creation:
                assert file_edit.previous_file_lines is not None, "Missing previous lines"
                for line in range(a, a + b):
                    diff_lines.append(f"-{file_edit.previous_file_lines[line - 1]}")
            for line in replacement.new_lines:
                diff_lines.append(f"+{line}")

        return "\n".join(diff_lines)

    def file_edits_to_llm_message(self, parsedLLMResponse: ParsedLLMResponse) -> str:
        ans = parsedLLMResponse.conversation.strip() + "\n\n"
        for file_edit in parsedLLMResponse.file_edits:
            ans += self.file_edit_to_git_diff(file_edit) + "\n"

        return ans



import sys

# check must be located before importing code that uses 3.10 features
if sys.version_info < (3, 10):
    exit("Error: Python version 3.10 or higher is required.")


from mentat.terminal.client import run_cli

# This is only here so that mentat can be run with python -m mentat
# This file will NOT run when run from pip installation
run_cli()



from typing import List

import attr
from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class ConfigCommand(Command, command_name="config"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config
        if len(args) == 0:
            stream.send("No config option specified", style="warning")
        else:
            setting = args[0]
            if hasattr(config, setting):
                if len(args) == 1:
                    value = getattr(config, setting)
                    description = attr.fields_dict(type(config))[setting].metadata.get("description")
                    stream.send(f"{setting}: {value}")
                    if description:
                        stream.send(f"Description: {description}")
                elif len(args) == 2:
                    value = args[1]
                    if attr.fields_dict(type(config))[setting].metadata.get("no_midsession_change"):
                        stream.send(
                            f"Cannot change {setting} mid-session. Please restart" " Mentat to change this setting.",
                            style="warning",
                        )
                        return
                    try:
                        setattr(config, setting, value)
                        stream.send(f"{setting} set to {value}", style="success")
                    except (TypeError, ValueError):
                        stream.send(f"Illegal value for {setting}: {value}", style="error")
                else:
                    stream.send("Too many arguments", style="warning")
            else:
                stream.send(f"Unrecognized config option: {setting}", style="error")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [
            CommandArgument("required", "setting"),
            CommandArgument("optional", "value"),
        ]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        # Dodge circular imports
        from mentat.config import Config

        if argument_position == 0:
            return Config.get_fields()
        elif argument_position == 1:
            setting = arguments[0]
            fields = attr.fields_dict(Config)
            if setting in fields:
                return fields[setting].metadata.get("auto_completions", [])
            else:
                return []
        else:
            return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Show or set a config option's value."



import json
import os
import webbrowser
from pathlib import Path
from typing import Optional, Tuple

import attr
from jinja2 import Environment, FileSystemLoader, select_autoescape

from benchmarks.benchmark_result import BenchmarkResult
from benchmarks.benchmark_run_summary import BenchmarkRunSummary


class BenchmarkRun:
    def __init__(
        self,
        results: list[BenchmarkResult],
        metadata: Optional[dict] = None,
    ):
        self.results = results
        self.metadata = metadata or {}
        self.results_map = {result.name: result for result in results}
        self.result_groups = self.group_results()
        self.summary: BenchmarkRunSummary = self.aggregate_results()

    def group_results(self) -> dict[str, list[BenchmarkResult]]:
        groups = {}
        for result in self.results:
            if result.family is None:
                groups[result.name] = [result]
            else:
                if result.family not in groups:
                    groups[result.family] = []
                groups[result.family].append(result)
        return groups

    def aggregate_results(self) -> BenchmarkRunSummary:
        summary: dict[str, Tuple[int | float | str, float]] = {}
        for field in attr.fields(BenchmarkResult):
            if "aggregation" in field.metadata:
                name = field.name
                values = [getattr(result, name) for result in self.results if getattr(result, name) is not None]
                if len(values) == 0:
                    summary[name] = (0, 0)
                else:
                    total_set = len(values)
                    aggregation_type = field.metadata["aggregation"]
                    if aggregation_type == "sum":
                        summary[name] = (sum(values), total_set)
                    elif aggregation_type == "average":
                        summary[name] = (sum(values) / len(values), total_set)
                    elif aggregation_type == "percent":
                        summary[name] = (sum(values) / len(values) * 100, total_set)
                    elif aggregation_type == "histogram":
                        histogram = {val: values.count(val) for val in set(values)}
                        summary[name] = (histogram, len(values))
                    elif aggregation_type == "none":
                        summary[name] = (values, len(values))
        return BenchmarkRunSummary(summary, self.metadata)

    def formatted_results(self) -> list[dict[str, str]]:
        formatted = {}
        for result in self.results:
            formatted_result = {}
            for field in attr.fields(BenchmarkResult):
                if "display" in field.metadata:
                    name = field.name
                    value = getattr(result, name)
                    if value is not None:
                        display_name = field.metadata.get("display_name", name)
                        formatted_result[display_name] = {
                            "content": value,
                            "type": field.metadata["display"],
                        }
            formatted[result.name] = formatted_result
        return formatted

    def make_html_report(self, output_path: Path = Path("results.html")):
        env = Environment(
            loader=FileSystemLoader(
                [
                    os.path.join(os.path.dirname(__file__), "../mentat/resources/templates"),
                    os.path.join(os.path.dirname(__file__), "resources/templates"),
                ]
            ),
            autoescape=select_autoescape(["html", "xml"]),
        )
        template = env.get_template("benchmark.jinja")
        rendered_html = template.render(benchmark_run=self)

        with open(output_path, "w") as f:
            f.write(rendered_html)

    def render_results(self):
        self.make_html_report()

        webbrowser.open("file://" + os.path.realpath("results.html"))

    def to_json(self) -> str:
        return json.dumps(
            {
                "results": [result.to_dict() for result in self.results],
                "metadata": self.metadata,
            },
            indent=4,
        )

    def save(
        self,
        folder: Path = Path("."),
        name: str = "results.json",
        summary_dir: str = "summary",
    ):
        folder.mkdir(parents=True, exist_ok=True)
        summary_dir = folder / summary_dir
        summary_dir.mkdir(parents=True, exist_ok=True)
        file_path = folder / name
        summary_path = summary_dir / name
        with open(file_path, "w") as f:
            f.write(self.to_json())
        with open(summary_path, "w") as f:
            f.write(self.summary.to_json())

    @classmethod
    def load_json(cls, json_str: str) -> "BenchmarkRun":
        data = json.loads(json_str)
        results = [BenchmarkResult.from_dict(result) for result in data["results"]]
        metadata = data.get("metadata")
        return cls(results, metadata=metadata)

    @classmethod
    def load_file(cls, file: Path) -> "BenchmarkRun":
        file_name = os.path.basename(file)
        with open(file, "r") as f:
            run = cls.load_json(f.read())
        run.metadata["file"] = file_name  # Used for links
        return run



from __future__ import annotations

import asyncio
import logging
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Optional

import attr

from mentat.ctags import get_ctag_lines_and_names
from mentat.diff_context import annotate_file_message, parse_diff
from mentat.errors import MentatError
from mentat.git_handler import get_diff_for_file
from mentat.interval import INTERVAL_FILE_END, Interval
from mentat.llm_api_handler import count_tokens
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path

MIN_INTERVAL_LINES = 10


def split_file_into_intervals(
    feature: CodeFeature,
    min_lines: int | None = None,
) -> list[CodeFeature]:
    min_lines = min_lines or MIN_INTERVAL_LINES
    session_context = SESSION_CONTEXT.get()
    code_file_manager = session_context.code_file_manager
    n_lines = len(code_file_manager.read_file(feature.path))

    lines_and_names = get_ctag_lines_and_names(session_context.cwd.joinpath(feature.path))

    if len(lines_and_names) == 0:
        return [feature]

    lines, names = map(list, zip(*sorted(lines_and_names)))
    lines[0] = 1  # first interval covers from start of file
    draft_named_intervals = [(name, start, end) for name, start, end in zip(names, lines, lines[1:] + [n_lines])]

    def length(interval: tuple[str, int, int]):
        return interval[2] - interval[1]

    def merge_intervals(int1: tuple[str, int, int], int2: tuple[str, int, int]):
        return (f"{int1[0]},{int2[0]}", int1[1], int2[2])

    named_intervals = [draft_named_intervals[0]]
    for next_interval in draft_named_intervals[1:]:
        last_interval = named_intervals[-1]
        if length(last_interval) < min_lines:
            named_intervals[-1] = merge_intervals(last_interval, next_interval)
        elif length(next_interval) < min_lines and next_interval == draft_named_intervals[-1]:
            # this is the last interval it's too short, so merge it with previous
            named_intervals[-1] = merge_intervals(last_interval, next_interval)
        else:
            named_intervals.append(next_interval)

    if len(named_intervals) <= 1:
        return [feature]

    # Create and return separate features for each interval
    _features = list[CodeFeature]()
    for name, start, end in named_intervals:
        _feature = CodeFeature(
            feature.path,
            interval=Interval(start, end),
            name=name,
        )
        _features.append(_feature)
    return _features


@attr.define(frozen=True)
class CodeFeature:
    """
    Represents a section of the code_message which is included with the prompt.
    Includes a section of code and an annotation method.

    Attributes:
        path: The absolute path to the file.
        interval: The lines in the file.
        name: The names of the features/functions in this CodeFeature
    """

    path: Path = attr.field()
    interval: Interval = attr.field(factory=lambda: Interval(1, INTERVAL_FILE_END))
    # eq is set to false so that we can compare duplicate features without names getting in the way
    name: Optional[str] = attr.field(default=None, eq=False)

    def __attrs_post_init__(self):
        if not self.path.is_absolute():
            raise MentatError("CodeFeature path must be absolute.")

    def __repr__(self):
        return (
            f"CodeFeature(path={self.path}," f" interval={self.interval.start}-{self.interval.end}, name={self.name})"
        )

    def rel_path(self, cwd: Optional[Path] = None) -> str:
        if cwd is not None:
            path_string = str(get_relative_path(self.path, cwd))
        else:
            path_string = str(self.path)
        return path_string

    def interval_string(self) -> str:
        if not self.interval.whole_file():
            interval_string = f":{self.interval.start}-{self.interval.end}"
        else:
            interval_string = ""
        return interval_string

    def __str__(self, cwd: Optional[Path] = None) -> str:
        return self.rel_path(cwd) + self.interval_string()

    def get_code_message(self, standalone: bool = True) -> list[str]:
        """
        Gets this code features code message.
        If standalone is true, will include the filename at top and extra newline at the end.
        If feature contains entire file, will add inline diff annotations; otherwise, will append them to the end.
        """
        if not self.path.exists() or self.path.is_dir():
            return []

        session_context = SESSION_CONTEXT.get()
        code_file_manager = session_context.code_file_manager
        parser = session_context.config.parser
        code_context = session_context.code_context

        code_message: list[str] = []

        if standalone:
            # We always want to give GPT posix paths
            code_message_path = get_relative_path(self.path, session_context.cwd)
            code_message.append(str(code_message_path.as_posix()))

        # Get file lines
        file_lines = code_file_manager.read_file(self.path)
        for i, line in enumerate(file_lines):
            if self.interval.contains(i + 1):
                if parser.provide_line_numbers():
                    code_message.append(f"{i + parser.line_number_starting_index()}:{line}")
                else:
                    code_message.append(f"{line}")

        if standalone:
            code_message.append("")

        if self.path in code_context.diff_context.diff_files():
            diff = get_diff_for_file(code_context.diff_context.target, self.path)
            diff_annotations = parse_diff(diff)
            if self.interval.whole_file():
                code_message = annotate_file_message(code_message, diff_annotations)
            else:
                for section in diff_annotations:
                    # TODO: Place diff_annotations inside interval where they belong
                    if section.start >= self.interval.start and section.start < self.interval.end:
                        code_message += section.message
        return code_message

    def get_checksum(self) -> str:
        # TODO: Only update checksum if last modified time of file updates to conserve file system reads
        session_context = SESSION_CONTEXT.get()
        code_file_manager = session_context.code_file_manager

        return code_file_manager.get_file_checksum(self.path, self.interval)

    def count_tokens(self, model: str) -> int:
        code_message = self.get_code_message()
        return count_tokens("\n".join(code_message), model, full_message=False)


async def count_feature_tokens(features: list[CodeFeature], model: str) -> list[int]:
    """Return the number of tokens in each feature."""
    sem = asyncio.Semaphore(10)

    async def _count_tokens(feature: CodeFeature) -> int:
        async with sem:
            return feature.count_tokens(model)

    tasks = [_count_tokens(f) for f in features]
    return await asyncio.gather(*tasks)


def _get_code_message_from_intervals(features: list[CodeFeature]) -> list[str]:
    """
    Merge multiple features for the same file into a single code message.
    """
    features_sorted = sorted(features, key=lambda f: f.interval)
    posix_path = features_sorted[0].get_code_message()[0]
    code_message = [posix_path]
    next_line = 1
    for feature in features_sorted:
        starting_line = feature.interval.start
        if starting_line < next_line:
            logging.info(f"Features overlap: {feature}")
            if feature.interval.end <= next_line:
                continue
            feature = CodeFeature(
                feature.path,
                interval=Interval(next_line, feature.interval.end),
                name=feature.name,
            )
        elif starting_line > next_line:
            code_message += ["..."]
        code_message += feature.get_code_message(standalone=False)
        next_line = feature.interval.end
    return code_message + [""]


def get_code_message_from_features(features: list[CodeFeature]) -> list[str]:
    """
    Generate a code message from a list of features.
    Will automatically handle overlapping intervals.
    """
    code_message = list[str]()
    features_by_path: dict[Path, list[CodeFeature]] = OrderedDict()
    for feature in features:
        if feature.path not in features_by_path:
            features_by_path[feature.path] = list[CodeFeature]()
        features_by_path[feature.path].append(feature)
    for path_features in features_by_path.values():
        if len(path_features) == 1:
            code_message += path_features[0].get_code_message()
        else:
            code_message += _get_code_message_from_intervals(path_features)
    return code_message


def get_consolidated_feature_refs(features: list[CodeFeature]) -> list[str]:
    """
    Return a list of 'path:<interval>,<interval>' strings, merging code features with the same path
    """
    level_info_by_path = defaultdict[Path, list[Interval | None]](list)
    for f in features:
        if f.interval.whole_file():
            level_info_by_path[f.path].append(None)
        else:
            level_info_by_path[f.path].append(f.interval)

    consolidated_refs = list[str]()
    for path, level_info in level_info_by_path.items():
        ref_string = str(path)
        intervals = sorted([level for level in level_info if level is not None])
        if intervals and None not in level_info:
            ref_string += ":" + ",".join(str(interval) for interval in intervals)
        consolidated_refs.append(ref_string)

    return consolidated_refs



#!/usr/bin/env python
import asyncio
import json
import os
import subprocess
from itertools import islice
from pathlib import Path
from textwrap import dedent

from git import Repo
from openai import OpenAI

from benchmarks.arg_parser import common_benchmark_parser
from mentat import Mentat
from mentat.sampler.utils import clone_repo


def load_tests(benchmarks_dir):
    tests = {}
    benchmarks_path = benchmarks_dir / "benchmarks.json"
    if benchmarks_path.exists():
        with open(benchmarks_path, "r") as f:
            tests = json.load(f)
    return tests


def load_results(benchmarks_dir):
    results = {}
    results_path = benchmarks_dir / "benchmark_results.json"
    if results_path.exists():
        with open(results_path, "r") as f:
            results = json.load(f)
    return results


def write_result(commit, result, repo_path):
    results = load_results(repo_path)
    results[commit] = result
    with open(repo_path / "benchmark_results.json", "w") as f:
        json.dump(results, f, indent=4)


grader_prompt = dedent(
    """\
        Please grade the following diff on the following metrics:
        - correctness
        - readability
        - style
        - surprisingness
        Please reply with only a json object rating the diff from 1
        to 5 on each of those dimensions. For example:
        {"correctness": 4, "readability": 3, "style": 2, "surprisingness": 1}
        """
)


def evaluate_diff(diff: str) -> dict[str, int]:
    messages = [
        {"role": "system", "content": grader_prompt},
        {"role": "system", "content": diff},
    ]

    client = OpenAI()
    response = client.chat.completions.create(
        model="gpt-4-0314",
        messages=messages,
    )
    message = response.choices[0].message.content

    return json.loads(message)


async def test_edit_quality(benchmarks, max_benchmarks, evaluate_baseline, repo, refresh_repo):
    repo_path = Path(__file__).parent / f"../../benchmark_repos/{repo}"
    tests = load_tests(repo_path)
    results = load_results(repo_path)

    if len(benchmarks) > 0:
        tests_to_run = {k: v for k, v in tests.items() if k in benchmarks}
    else:
        tests_to_run = dict(islice(tests.items(), max_benchmarks))

    for test in tests_to_run.values():
        print(f"\n\n{test['name']}\n{test['prompt']}")
        if test["commit"] in results and not refresh_repo:
            print("Already ran")
            print(results[test["commit"]])
            continue

        repo_url = test["codebase_url"]
        repo_name = repo_url.split("/")[-1]
        codebase = clone_repo(repo_url, repo_name)
        os.chdir(codebase)
        with open(".git/info/exclude", "w") as f:
            f.write(
                dedent(
                    """\
                commit_information.json
                benchmarks.json
                benchmark_results.json
                transcripts*.jsonl
                gpt-output-cache.json
                """
                )
            )
        repo = Repo(".")
        start_commit = repo.commit()
        repo.git.checkout(test["commit"] + "^1")

        client = Mentat(
            paths=test["expected_features"],
        )

        await client.startup()
        await client.call_mentat_auto_accept(test["prompt"])
        await client.wait_for_edit_completion()
        await client.call_mentat("/commit")
        # I don't think this should be necessary but without it the diff isn't
        # ready in the next line.
        await client.call_mentat("q")

        diff = subprocess.check_output(["git", "show"]).decode("utf-8")
        evaluation = evaluate_diff(diff)

        print("Mentat produced the following diff:")
        print(diff)
        print("And GPT rates it in the following way:")
        print(evaluation)
        result = {
            "diff": diff,
            "grade": evaluation,
        }

        if evaluate_baseline:
            baseline = test["expected_edits"]
            evaluated_baseline = evaluate_diff(baseline)
            result["baseline"] = evaluated_baseline
            print("GPT thinks the actual diff rates:")
            print(evaluated_baseline)

        write_result(test["commit"], result, repo_path)

        # Clean up
        repo.git.reset("--hard")
        repo.git.clean("-fd")
        repo.git.checkout(start_commit)
        await client.shutdown()


if __name__ == "__main__":
    parser = common_benchmark_parser()
    args = parser.parse_args()
    asyncio.run(
        test_edit_quality(
            args.benchmarks,
            args.max_benchmarks,
            args.evaluate_baseline,
            args.repo,
            args.refresh_repo,
        )
    )



import webbrowser
from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT
from mentat.transcripts import Transcript, get_transcript_logs
from mentat.utils import create_viewer


class ViewerCommand(Command, command_name="viewer"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        conversation = session_context.conversation

        logs = get_transcript_logs()

        viewer_path = create_viewer([Transcript(id="Current", messages=conversation.literal_messages)] + logs)
        webbrowser.open(f"file://{viewer_path.resolve()}")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Open a webpage showing the conversation so far."



import plotly.graph_objs as go
from plotly.offline import plot

# An array of filters to trace property
filters = [
    (
        (
            lambda x: x.metadata["branch"] == "main"
            and x.metadata["type"] == "exercism"
            and x.metadata["language"] == "python"
        ),
        {
            "mode": "lines+markers",
            "name": "main python exercism",
            "line": {"color": "green"},
            "marker": {"color": "green"},
        },
    ),
    (
        (
            lambda x: x.metadata["branch"] != "main"
            and x.metadata["type"] == "exercism"
            and x.metadata["language"] == "python"
        ),
        {
            "mode": "markers",
            "name": "python exercism experiments",
            "marker": {"color": "green"},
        },
    ),
    (
        (
            lambda x: x.metadata["branch"] == "main"
            and x.metadata["type"] == "exercism"
            and x.metadata["language"] == "javascript"
        ),
        {
            "mode": "lines+markers",
            "name": "main javascript exercism",
            "line": {"color": "red"},
            "marker": {"color": "red"},
        },
    ),
    (
        (
            lambda x: x.metadata["branch"] != "main"
            and x.metadata["type"] == "exercism"
            and x.metadata["language"] == "javascript"
        ),
        {
            "mode": "markers",
            "name": "javascript exercism experiments",
            "marker": {"color": "red"},
        },
    ),
    (
        (lambda x: x.metadata["branch"] == "main" and x.metadata["type"] != "exercism"),
        {
            "mode": "lines+markers",
            "name": "main real world",
            "line": {"color": "black"},
            "marker": {"color": "black"},
        },
    ),
    (
        (lambda x: x.metadata["branch"] != "main" and x.metadata["type"] != "exercism"),
        {
            "mode": "markers",
            "name": "real world experiments",
            "marker": {"color": "black"},
        },
    ),
]


def generate_plot_html(summary_list):
    summary_list = sorted(summary_list, key=lambda x: x.metadata["date"])

    data = []
    for f, props in filters:
        data.append(
            go.Scatter(
                x=[summary.metadata["date"] for summary in summary_list if f(summary)],
                y=[summary.summary["passed"][0] for summary in summary_list if f(summary)],
                text=[summary.metadata["file"] for summary in summary_list if f(summary)],
                **props,
            )
        )

    # Generate buttons for each metric available in the summaries
    metrics = list(summary_list[0].summary.keys())
    buttons = []
    for metric in metrics:
        ys = []
        for f, props in filters:
            ys.append([summary.summary[metric][0] for summary in summary_list if f(summary)])
        buttons.append(dict(args=[{"y": ys}], label=metric, method="update"))

    # Add the buttons to the layout
    updatemenus = list(
        [
            dict(
                buttons=buttons,
                direction="down",
                pad={"r": 10, "t": 10},
                showactive=True,
                x=1,
                xanchor="right",
                y=0.9,
                active=metrics.index("passed"),
                yanchor="middle",
            ),
        ]
    )

    layout = go.Layout(title="Benchmarks", showlegend=True, updatemenus=updatemenus)

    fig = go.Figure(data=data, layout=layout)
    return plot(fig, output_type="div", include_plotlyjs=True)



from __future__ import annotations

import asyncio
import hashlib
import time
from importlib import resources
from importlib.abc import Traversable
from pathlib import Path
from typing import TYPE_CHECKING, AsyncIterator, List, Literal, Optional, Union

from jinja2 import Environment, PackageLoader, select_autoescape
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDelta

if TYPE_CHECKING:
    from mentat.transcripts import Transcript


mentat_dir_path = Path.home() / ".mentat"

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
resources_path = Path("resources")
conversation_viewer_path = Path("conversation_viewer.html")


def sha256(data: str) -> str:
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


async def run_subprocess_async(*args: str) -> str:
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        start_new_session=True,
    )
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        error_output = stderr.decode("utf-8").strip() if stderr else ""
        raise Exception(f"Subprocess failed with error: {error_output}")

    output = stdout.decode("utf-8").strip() if stdout else ""

    return output


# Useful for using functions designed to work with LLMs on prepared strings
async def convert_string_to_asynciter(
    input_str: str,
    chunk_size: int,
    role: Optional[Literal["system", "user", "assistant", "tool"]] = "assistant",
) -> AsyncIterator[ChatCompletionChunk]:
    timestamp = int(time.time())
    for i in range(0, len(input_str), chunk_size):
        yield ChatCompletionChunk(
            id="asynciter-id",
            choices=[
                Choice(
                    delta=ChoiceDelta(content=input_str[i : i + chunk_size], role=role),
                    finish_reason=None,
                    index=0,
                )
            ],
            created=timestamp,
            model="asynciter-model",
            object="chat.completion.chunk",
        )


def fetch_resource(resource_path: Path) -> Traversable:
    resource = resources.files(package_name).joinpath(str(resources_path / resource_path))
    return resource


def create_viewer(transcripts: list[Transcript]) -> Path:
    env = Environment(
        loader=PackageLoader("mentat", "resources/templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
    template = env.get_template("conversation_viewer.jinja")
    html = template.render(transcripts=transcripts[:500])

    viewer_path = mentat_dir_path / conversation_viewer_path
    with viewer_path.open("w") as viewer_file:
        viewer_file.write(html)
    return viewer_path


async def add_newline(
    iterator: AsyncIterator[ChatCompletionChunk],
    role: Optional[Literal["system", "user", "assistant", "tool"]] = "assistant",
) -> AsyncIterator[ChatCompletionChunk]:
    """
    The model often doesn't end it's responses in a newline;
    adding a newline makes it significantly easier for us to parse.
    """
    last_chunk = None
    async for chunk in iterator:
        last_chunk = chunk
        yield chunk
    if last_chunk is not None:
        yield ChatCompletionChunk(
            id=last_chunk.id,
            choices=[
                Choice(
                    delta=ChoiceDelta(content="\n", role=role),
                    finish_reason=last_chunk.choices[0].finish_reason,
                    index=0,
                )
            ],
            created=last_chunk.created,
            model=last_chunk.model,
            object=last_chunk.object,
            system_fingerprint=last_chunk.system_fingerprint,
        )


def get_relative_path(path: Path, target: Path) -> Path:
    """Get the relative path of a file from a given directory.

    This function is a 'backport' of `PurePath.relative_to` from Python 3.12 and later which has directory walk-up
    support. See https://docs.python.org/3.12/library/pathlib.html#pathlib.PurePath.relative_to.

    Args:
        path (Path): The path to get the relative path of. This path must exist on the filesystem.
        target (Path): The target path to base the relative path off of. This path must exist on the filesystem.

    Returns:
        Path: A relative path
    """
    path = path.resolve()
    target = target.resolve()

    if path.is_relative_to(target):
        relative_path = path.relative_to(target)
    else:
        relative_parts: List[Union[str, Path]] = []
        parent = target
        while not path.is_relative_to(parent):
            relative_parts.append("..")
            parent = parent.parent
        relative_parts.append(path.relative_to(parent))
        relative_path = Path(*relative_parts)

    return relative_path


# TODO: replace this with something that doesn't load the file into memory
def is_file_text_encoded(abs_path: Path):
    """Checks if a file is text encoded."""
    try:
        # The ultimate filetype test
        with open(abs_path, "r") as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False



def matching_index(orig_lines: list[str], new_lines: list[str]) -> int:
    orig_lines = orig_lines.copy()
    new_lines = new_lines.copy()
    index = _exact_match(orig_lines, new_lines)
    if index == -1:
        orig_lines = [s.lower() for s in orig_lines]
        new_lines = [s.lower() for s in new_lines]
        index = _exact_match(orig_lines, new_lines)
        if index == -1:
            orig_lines = [s.strip() for s in orig_lines]
            new_lines = [s.strip() for s in new_lines]
            index = _exact_match(orig_lines, new_lines)
            if index == -1:
                new_orig_lines = [s for s in orig_lines if s]
                new_new_lines = [s for s in new_lines if s]
                index = _exact_match(new_orig_lines, new_new_lines)
                if new_orig_lines and index != -1:
                    index = orig_lines.index(new_orig_lines[index])
    return index


def _exact_match(orig_lines: list[str], new_lines: list[str]) -> int:
    if "".join(new_lines).strip() == "" and "".join(orig_lines).strip() == "":
        return 0
    for i in range(len(orig_lines) - (len(new_lines) - 1)):
        if orig_lines[i : i + len(new_lines)] == new_lines:
            return i
    return -1



import base64
import os
from typing import Optional

import attr
from selenium import webdriver
from selenium.common.exceptions import NoSuchWindowException, WebDriverException
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.edge.service import Service as EdgeService
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.remote.webdriver import WebDriver
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager

from mentat.session_context import SESSION_CONTEXT


class ScreenshotException(Exception):
    """
    Thrown when a screenshot is attempted before the browser is opened.
    """


@attr.define
class VisionManager:
    driver: Optional[WebDriver] = attr.field(default=None)

    def _open_browser(self) -> None:
        ctx = SESSION_CONTEXT.get()
        safari_installed = False
        if self.driver is None or not self.driver_running():
            try:
                self.driver = webdriver.Safari()
            except Exception as e:
                if "remote automation" in str(e).lower():
                    safari_installed = True
                try:
                    service = Service(ChromeDriverManager().install())
                    self.driver = webdriver.Chrome(service=service)
                except Exception:
                    try:
                        service = EdgeService(EdgeChromiumDriverManager().install())
                        self.driver = webdriver.Edge(service=service)
                    except Exception:
                        try:
                            service = FirefoxService(GeckoDriverManager().install())
                            self.driver = webdriver.Firefox(service=service)
                        except Exception:
                            if safari_installed:
                                ctx.stream.send(
                                    "No suitable browser found. To use Safari, enable" " remote automation.",
                                    style="error",
                                )
                            else:
                                ctx.stream.send(
                                    "No suitable browser found.",
                                    style="error",
                                )

                            raise ScreenshotException()

    def open(self, path: str) -> None:
        self._open_browser()
        self.driver.get(path)  # type: ignore

    def driver_running(self):
        try:
            # This command should fail if the driver is not running
            self.driver.execute_script('return "hello world";')  # type: ignore
            return True
        except NoSuchWindowException:
            return False

    def screenshot(self, path: Optional[str] = None) -> str:
        ctx = SESSION_CONTEXT.get()
        if path is not None:
            expanded = os.path.abspath(os.path.expanduser(path))
            browser_path = path
            if os.path.exists(expanded):
                browser_path = "file://" + expanded
            else:
                if not path.startswith("http"):
                    browser_path = "https://" + path
            try:
                self.open(browser_path)
            except WebDriverException:
                ctx.stream.send(
                    f"Error taking screenshot. Is {path} a valid url or local file?",
                    style="error",
                )
                raise ScreenshotException()
        else:
            if self.driver is None:
                ctx.stream.send(
                    'No browser open. Run "/screenshot path" with a url or local file',
                    style="error",
                )
                raise ScreenshotException()

        screenshot_data = self.driver.get_screenshot_as_png()  # type: ignore

        decoded = base64.b64encode(screenshot_data).decode("utf-8")
        image_data = f"data:image/png;base64,{decoded}"

        return image_data

    def close(self) -> None:
        if self.driver is not None:
            self.driver.quit()



import ast
import re
from pathlib import Path
from textwrap import dedent
from typing import List, Optional, Tuple

import openai
from watchfiles import Change, watch

openai_client = openai.Client()


def get_mentat_comment(filename: Path) -> Optional[Tuple[int, str]]:
    with open(filename, "r") as file:
        for line_number, line in enumerate(file, 1):
            if line.startswith("# @mentat"):
                comment_text = line.split("@mentat", 1)[1].strip()
                return line_number, comment_text
    return None


prompt = """
    You are an expert coding assistant. 

    You will be given a code file and a coding task to complete. Your response will be inserted into 
    the code file. Your response must be valid python code and must adhere strictly to the required 
    format. Put all of your response that's valid python code between '@@start' and '@@end' lines.
    Do not include any examples or instructions in your response. Do not re-write the entire file, 
    just answer the coding task.

    Example Request:

    Code File:
    ```
    def add_numbers(a, b):
        return a + b


    def multiply_numbers(a, b):
        return a * b


    def subtract_numbers(a, b):
        return a - b
    ```

    Code Task: create a new function that divides 2 numbers together.


    Example Response:
    
    @@start
    def divide_numbers(a, b):
        return a / b
    @@end
"""


def generate_code_lines(*, code_file: Path, code_task: str) -> List[str]:
    user_prompt = f"""
        Code File:
        ```
        {code_file.read_text()} 
        ```

        Code Task: {code_task}
    """
    user_prompt = dedent(user_prompt).strip()

    # Call openai
    chat = openai_client.chat.completions.create(
        messages=[
            {"role": "system", "content": dedent(prompt.strip())},
            {"role": "user", "content": user_prompt},
        ],
        model="gpt-4-0125-preview",
    )
    result = chat.choices[0].message.content
    if result is None:
        return []

    # Verify code is valid python (else re-run?)
    pattern = r"@@start(.*?)@@end"
    matches = re.findall(pattern, result, re.DOTALL)
    if len(matches) == 0:
        return []

    match = matches[0]
    try:
        # Parse the code to verify its validity
        ast.parse(match)
    except SyntaxError:
        return []

    return match.splitlines()


def modify_file_lines(*, path: Path, user_prompt: str, code_lines: List[str]):
    with open(path, "r") as file:
        file_lines = file.readlines()

    # Find the line number the user_prompt is on
    line_number = None
    for _line_number, line in enumerate(file_lines, 1):
        if line.strip() == f"# @mentat {user_prompt}":
            line_number = _line_number
    if line_number is None:
        print("Could not find the line number for the user prompt")
        return

    # Modify the comment
    file_lines.pop(line_number - 1)
    file_lines.insert(line_number - 1, f"# [completed] @mentat {user_prompt}\n")

    # Insert the code snippet into the file
    code_line_number = line_number
    for code_line in code_lines:
        if code_line == "":
            continue
        file_lines.insert(code_line_number, f"{code_line}\n")
        code_line_number += 1

    with open(path, "w") as file:
        file.writelines(file_lines)


def process_file_change(path: Path):
    # Get @mentat comment
    comment = get_mentat_comment(path)
    if comment is None:
        return
    comment_line, comment_text = comment
    print(f"{path}:{comment_line} >> {comment_text}")

    # Call openai
    print("Calling openai...")
    code_lines = generate_code_lines(code_file=path, code_task=comment_text)
    if len(code_lines) == 0:
        print("No code snippets generated")
        return
    else:
        print("Code snippets generated:", "\n".join(code_lines))

    # Modify the file
    print("Modifying file...")
    modify_file_lines(path=path, user_prompt=comment_text, code_lines=code_lines)
    print("Done!")


def main(path: Path = Path.cwd()):
    print(f"Watching {path}")
    for changes in watch(path, force_polling=True):
        for change_type, _change_path in changes:
            change_path = Path(_change_path)
            if change_path.is_file() is False:
                continue
            if change_type == Change.modified:
                process_file_change(change_path)



import fnmatch
import glob
import os
import re
from enum import Enum
from pathlib import Path
from typing import List, Set

from mentat.code_feature import CodeFeature
from mentat.errors import PathValidationError
from mentat.git_handler import get_git_root_for_path, get_non_gitignored_files
from mentat.interval import parse_intervals, split_intervals_from_path
from mentat.utils import is_file_text_encoded


class PathType(Enum):
    FILE = "file"
    FILE_INTERVAL = "file_interval"
    DIRECTORY = "directory"
    GLOB = "glob"


def is_interval_path(path: Path) -> bool:
    path, interval_str = split_intervals_from_path(path)
    if not interval_str:
        return False
    intervals = parse_intervals(interval_str)
    if len(intervals) == 0:
        return False
    return True


def get_path_type(path: Path) -> PathType:
    """Get the type of path.

    Args:
        `path` - An absolute path

    Return:
        A PathType enum
    """
    if not path.is_absolute():
        raise PathValidationError(f"Path {path} is not absolute")

    if path.is_file():
        return PathType.FILE
    elif is_interval_path(path):
        return PathType.FILE_INTERVAL
    elif path.is_dir():
        return PathType.DIRECTORY
    elif re.search(r"[\*\?\[\]]", str(path)):
        return PathType.GLOB
    else:
        raise PathValidationError(f"Path {path} does not exist")


def validate_file_path(path: Path, check_for_text: bool = True) -> None:
    if not path.is_absolute():
        raise PathValidationError(f"File {path} is not absolute")
    if not path.exists():
        raise PathValidationError(f"File {path} does not exist")
    if check_for_text and not is_file_text_encoded(path):
        raise PathValidationError(f"Unable to read file {path}")


def validate_file_interval_path(path: Path, check_for_text: bool = True) -> None:
    interval_path, interval_str = split_intervals_from_path(path)
    if not interval_path.is_absolute():
        raise PathValidationError(f"File interval {path} is not absolute")
    if not interval_path.exists():
        raise PathValidationError(f"File {interval_path} does not exist")
    if check_for_text and not is_file_text_encoded(interval_path):
        raise PathValidationError(f"Unable to read file {interval_path}")

    # check that there is at least one interval
    intervals = parse_intervals(interval_str)
    if len(intervals) == 0:
        raise PathValidationError(f"Unable to parse intervals for path {interval_path}")

    # check that each interval exists
    if check_for_text:
        with open(interval_path, "r") as f:
            line_count = len(f.read().split("\n"))
        for interval in intervals:
            if interval.start < 0 or interval.end > line_count + 1:
                raise PathValidationError(
                    f"Interval {interval.start}-{interval.end} is out of bounds for" f" file {interval_path}"
                )


def validate_glob_path(path: Path) -> None:
    if not path.is_absolute():
        raise PathValidationError(f"Glob path {path} is not absolute")
    try:
        glob.iglob(str(path)).__next__()
    except StopIteration:
        raise PathValidationError(f"Unable to validate glob path {path}")


def validate_and_format_path(path: Path | str, cwd: Path, check_for_text: bool = True) -> Path:
    """Validate and format a path.

    Args:
        `path` - A file path, file interval path, directory path, or a glob pattern
        `check_for_text` - Check if the file can be opened. Default to True

    Return:
        An absolute path
    """
    path = Path(path)

    # Resolve ~
    try:
        path = path.expanduser()
    except RuntimeError as e:
        raise PathValidationError(f"Error expanding path {path}: {e}")

    # Get absolute path
    if path.is_absolute():
        abs_path = path
    else:
        abs_path = cwd / path

    # Resolve path (remove any '..' or symlinks)
    abs_path = abs_path.resolve()

    # Validate path
    match get_path_type(abs_path):
        case PathType.FILE:
            validate_file_path(abs_path, check_for_text)
        case PathType.FILE_INTERVAL:
            validate_file_interval_path(abs_path, check_for_text)
        case PathType.DIRECTORY:
            pass
        case PathType.GLOB:
            validate_glob_path(abs_path)

    return abs_path


def match_path_with_patterns(path: Path, patterns: Set[Path]) -> bool:
    """Check if the given absolute path matches any of the patterns.

    Args:
        `path` - An absolute path
        `patterns` - A set of absolute paths/glob patterns

    Return:
        A boolean flag indicating if the path matches any of the patterns
    """
    if not path.is_absolute():
        raise PathValidationError(f"Path {path} is not absolute")
    for pattern in patterns:
        if not pattern.is_absolute():
            raise PathValidationError(f"Pattern {pattern} is not absolute")
        # Check if the path is relative to the pattern
        if path.is_relative_to(pattern):
            return True
        # Check if the pattern is a glob pattern match
        if fnmatch.fnmatch(str(path), str(pattern)):
            return True
    return False


def get_paths_for_directory(
    path: Path,
    include_patterns: Set[Path] = set(),
    exclude_patterns: Set[Path] = set(),
    recursive: bool = True,
) -> Set[Path]:
    """Get all file paths in a directory.

    Args:
        `path` - An absolute path to a directory on the filesystem
        `include_patterns` - An iterable of absolute paths/glob patterns to include
        `exclude_patterns` - An iterable of absolute paths/glob patterns to exclude
        `recursive` - A boolean flag to recursive traverse child directories

    Return:
        A set of absolute file paths
    """
    paths: Set[Path] = set()

    if not path.exists():
        raise PathValidationError(f"Path {path} does not exist")
    if not path.is_dir():
        raise PathValidationError(f"Path {path} is not a directory")
    if not path.is_absolute():
        raise PathValidationError(f"Path {path} is not absolute")

    for root, dirs, files in os.walk(path, topdown=True):
        root = Path(root)

        if get_git_root_for_path(root, raise_error=False):
            dirs[:] = list[str]()
            git_non_gitignored_paths = get_non_gitignored_files(root)
            for git_path in git_non_gitignored_paths:
                abs_git_path = root / git_path
                if not recursive and git_path.parent != Path("."):
                    continue
                if any(include_patterns) and not match_path_with_patterns(abs_git_path, include_patterns):
                    continue
                if any(exclude_patterns) and match_path_with_patterns(abs_git_path, exclude_patterns):
                    continue
                paths.add(abs_git_path)

        else:
            filtered_dirs: List[str] = []
            for dir_ in dirs:
                abs_dir_path = root.joinpath(dir_)
                if any(include_patterns) and not match_path_with_patterns(abs_dir_path, include_patterns):
                    continue
                if any(exclude_patterns) and match_path_with_patterns(abs_dir_path, exclude_patterns):
                    continue
                filtered_dirs.append(dir_)
            dirs[:] = filtered_dirs

            for file in files:
                abs_file_path = root.joinpath(file)
                if any(include_patterns) and not match_path_with_patterns(abs_file_path, include_patterns):
                    continue
                if any(exclude_patterns) and match_path_with_patterns(abs_file_path, exclude_patterns):
                    continue
                paths.add(abs_file_path)

            if not recursive:
                break
    paths = set(p.resolve() for p in paths if is_file_text_encoded(p))

    return paths


def get_code_features_for_path(
    path: Path,
    cwd: Path,
    include_patterns: Set[Path] = set(),
    exclude_patterns: Set[Path] = set(),
) -> Set[CodeFeature]:
    validated_path = validate_and_format_path(path, cwd)

    match get_path_type(validated_path):
        case PathType.FILE:
            code_features = set([CodeFeature(validated_path)])
        case PathType.FILE_INTERVAL:
            interval_path, interval_str = split_intervals_from_path(validated_path)
            intervals = parse_intervals(interval_str)
            code_features: Set[CodeFeature] = set()
            for interval in intervals:
                code_feature = CodeFeature(interval_path, interval)
                code_features.add(code_feature)
        case PathType.DIRECTORY:
            paths = get_paths_for_directory(validated_path, include_patterns, exclude_patterns)
            code_features = set(CodeFeature(p) for p in paths)
        case PathType.GLOB:
            root_parts: List[str] = []
            pattern: str | None = None
            for i, part in enumerate(validated_path.parts):
                if re.search(r"[\*\?\[\]]", str(part)):
                    pattern = str(Path().joinpath(*validated_path.parts[i:]))
                    break
                root_parts.append(part)
            if pattern is None:
                raise PathValidationError(f"Unable to parse glob pattern {validated_path}")
            root = Path().joinpath(*root_parts)
            paths = get_paths_for_directory(
                root,
                include_patterns=(set([*include_patterns, validated_path]) if pattern != "*" else include_patterns),
                exclude_patterns=exclude_patterns,
                recursive=True,
            )
            code_features = set(CodeFeature(p) for p in paths)

    return code_features



from pathlib import Path
import subprocess
import re
from typing import Any

from mentat import Mentat
from mentat.errors import SampleError
from mentat.git_handler import get_git_diff
from mentat.parsers.git_parser import GitParser
from mentat.sampler.sample import Sample
from mentat.sampler.utils import get_active_snapshot_commit, setup_repo, apply_diff_to_repo
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import convert_string_to_asynciter


async def run_sample(sample: Sample, cwd: Path | str | None = None) -> dict[str, Any]:
    """Run a sample using Mentat and return the resulting diff"""

    repo = setup_repo(
        url=sample.repo,
        cwd=cwd,
        commit=sample.merge_base,
        diff_merge_base=sample.diff_merge_base,
        diff_active=sample.diff_active,
    )
    cwd = Path(repo.working_dir)

    # Make a commit from the pre-edited state (should match diff_active)
    commit_active = get_active_snapshot_commit(repo)

    # Run sample in PythonClient
    paths = list[Path]()
    for a in sample.context:
        paths.append(Path(a))
    mentat = Mentat(cwd=cwd, paths=paths)
    await mentat.startup()
    session_context = SESSION_CONTEXT.get()
    conversation = session_context.conversation
    cost_tracker = session_context.cost_tracker
    for msg in sample.message_history:
        if msg["role"] == "user":
            conversation.add_user_message(msg["content"])
        elif msg["role"] == "assistant":
            generator = convert_string_to_asynciter(msg["content"], 100)
            parsed_llm_response = await GitParser().stream_and_parse_llm_response(generator)
            content = session_context.config.parser.file_edits_to_llm_message(parsed_llm_response)
            conversation.add_model_message(content, [], parsed_llm_response)
        else:
            raise SampleError(f"Invalid role found in message_history: {msg['role']}")
    prompt = sample.message_prompt
    if sample.hint_text:
        prompt += f"\n{80 * '-'}\nHint Text:\n{sample.hint_text}"
    await mentat.call_mentat_auto_accept(prompt)
    await mentat.shutdown()

    # Get the diff between pre- and post-edit
    transcript_messages = conversation.literal_messages.copy()

    message_eval = str(transcript_messages[-1].get("message", ""))
    diff_eval = get_git_diff(commit_active or "HEAD", cwd=cwd)

    test_results = {"passed": 0, "failed": 0, "error": ""}
    if sample.test_command:
        if sample.test_patch:
            apply_diff_to_repo(sample.test_patch, repo)
        try:
            output = subprocess.run(
                sample.test_command,
                shell=True,
                capture_output=True,
                text=True,
                cwd=cwd,
            )
            matches = re.search(r"(?:(\d+) passed)?(?:, )?(?:(\d+) failed)?", output.stdout)
            if matches:
                test_results["passed"] = int(matches.group(1)) or 0
                test_results["failed"] = int(matches.group(2)) or 0
            else:
                raise SampleError(f"Test command failed: {output.stdout}")
        except Exception as e:
            test_results["error"] = str(e)

    return {
        "id": sample.id,
        "message_eval": message_eval,
        "diff_eval": diff_eval,
        "cost": cost_tracker.total_cost,
        "tokens": cost_tracker.total_tokens,
        "transcript": {
            "id": sample.id,
            "messages": transcript_messages,
        },
        "test_results": test_results,
    }



import json
from pathlib import Path
from timeit import default_timer
from typing import Optional, Set

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.completion_create_params import ResponseFormat

from mentat.code_feature import CodeFeature, get_code_message_from_features
from mentat.errors import ModelError, PathValidationError, UserError
from mentat.feature_filters.feature_filter import FeatureFilter
from mentat.feature_filters.truncate_filter import TruncateFilter
from mentat.include_files import get_code_features_for_path
from mentat.llm_api_handler import count_tokens, model_context_size, prompt_tokens
from mentat.prompts.prompts import read_prompt
from mentat.session_context import SESSION_CONTEXT


class LLMFeatureFilter(FeatureFilter):
    feature_selection_prompt_path = Path("feature_selection_prompt.txt")

    def __init__(
        self,
        max_tokens: int,
        user_prompt: Optional[str] = None,
        expected_edits: Optional[list[str]] = None,
    ):
        self.max_tokens = max_tokens
        self.user_prompt = user_prompt or ""
        self.expected_edits = expected_edits

    async def filter(
        self,
        features: list[CodeFeature],
    ) -> list[CodeFeature]:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config
        cost_tracker = session_context.cost_tracker
        llm_api_handler = session_context.llm_api_handler

        stream.send(None, channel="loading")

        # Preselect as many features as fit in the context window
        model = config.feature_selection_model
        context_size = model_context_size(model)
        if context_size is None:
            raise UserError("Unknown context size for feature selection model: " f"{config.feature_selection_model}")
        context_size = min(context_size, config.llm_feature_filter)
        system_prompt = read_prompt(self.feature_selection_prompt_path)
        system_prompt_tokens = count_tokens(system_prompt, config.feature_selection_model, full_message=True)
        user_prompt_tokens = count_tokens(self.user_prompt, model, full_message=True)
        expected_edits_tokens = (
            0 if not self.expected_edits else count_tokens("\n".join(self.expected_edits), model, full_message=True)
        )
        preselect_max_tokens = (
            context_size - system_prompt_tokens - user_prompt_tokens - expected_edits_tokens - config.token_buffer
        )
        truncate_filter = TruncateFilter(preselect_max_tokens, model)
        preselected_features = await truncate_filter.filter(features)

        # Ask the model to return only relevant features
        messages: list[ChatCompletionMessageParam] = [
            ChatCompletionSystemMessageParam(role="system", content=system_prompt),
            ChatCompletionSystemMessageParam(
                role="system",
                content="\n".join(["CODE FILES:"] + get_code_message_from_features(preselected_features)),
            ),
            ChatCompletionUserMessageParam(role="user", content=f"USER QUERY: {self.user_prompt}"),
        ]
        if self.expected_edits:
            messages.append(
                ChatCompletionAssistantMessageParam(role="assistant", content=f"Expected Edits:\n{self.expected_edits}")
            )
        messages.append(
            ChatCompletionSystemMessageParam(
                role="system",
                content=(
                    "Now, identify the CODE FILES that are relevant to answering the"
                    " USER QUERY, Return a dict of {path: reason} for each file you"
                    " identify as relevant. e.g. {'src/main.js': 'Create new file',"
                    " 'public/index.html': 'Import main.js'}"
                ),
            )
        )
        selected_refs = list[Path]()
        start_time = default_timer()
        llm_response = await llm_api_handler.call_llm_api(
            messages=messages,
            model=model,
            stream=False,
            response_format=ResponseFormat(type="json_object"),
        )
        message = (llm_response.choices[0].message.content) or ""
        tokens = prompt_tokens(messages, model)
        response_tokens = count_tokens(message, model, full_message=True)
        cost_tracker.log_api_call_stats(
            tokens,
            response_tokens,
            model,
            default_timer() - start_time,
        )
        stream.send(None, channel="loading")

        # Parse response into features
        try:
            response = json.loads(message)  # type: ignore
            selected_refs = [Path(r) for r in response]
        except json.JSONDecodeError:
            raise ModelError(f"The response is not valid json: {message}")
        postselected_features: Set[CodeFeature] = set()
        for selected_ref in selected_refs:
            try:
                parsed_features = get_code_features_for_path(path=selected_ref, cwd=session_context.cwd)
                for feature in parsed_features:
                    assert any(
                        in_feat.path == feature.path and in_feat.interval.intersects for in_feat in preselected_features
                    )
                    postselected_features.add(feature)
            except (PathValidationError, AssertionError):
                stream.send(
                    f"LLM selected invalid path: {selected_ref}, skipping.",
                    style="warning",
                )

        # Truncate again to enforce max_tokens
        truncate_filter = TruncateFilter(self.max_tokens, config.model)
        return await truncate_filter.filter(postselected_features)



import datetime
import logging

from mentat.llm_api_handler import is_test_environment
from mentat.utils import mentat_dir_path

logs_dir = "logs"
logs_path = mentat_dir_path / logs_dir


def setup_logging():
    if is_test_environment():
        return

    logging.getLogger("openai").setLevel(logging.WARNING)
    # Breaking out of async generator when model messes up causes an error
    logging.getLogger("asyncio").setLevel(logging.CRITICAL)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # Root logger
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    # Only log warnings and higher to console
    console_handler.setLevel(logging.WARNING)

    logs_path.mkdir(parents=True, exist_ok=True)

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    log_file = logs_path / f"mentat_{timestamp}.log"
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(formatter)

    handlers = [console_handler, file_handler]

    try:
        latest_log_file = logs_path / "latest.log"
        latest_log_file.unlink(missing_ok=True)

        file_handler_latest = logging.FileHandler(latest_log_file)
        file_handler_latest.setFormatter(formatter)
        handlers.append(file_handler_latest)
    except PermissionError:
        # Thrown on Windows when trying to unlink a file that's in use by another mentat process;
        # instead, we just run without a latest.log handler if mentat is already running.
        pass

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    for handler in root.handlers[:]:
        root.removeHandler(handler)
        handler.close()
    for handler in handlers:
        root.addHandler(handler)

    # Costs logger
    costs_logger = logging.getLogger("costs")
    for handler in costs_logger.handlers[:]:
        costs_logger.removeHandler(handler)
        handler.close()
    costs_formatter = logging.Formatter("%(asctime)s\n%(message)s")
    costs_handler = logging.FileHandler(logs_path / "costs.log")
    costs_handler.setFormatter(costs_formatter)
    costs_logger.addHandler(costs_handler)
    costs_logger.setLevel(logging.INFO)
    costs_logger.propagate = False

    # Transcript logger
    transcripts_logger = logging.getLogger("transcript")
    for handler in transcripts_logger.handlers[:]:
        transcripts_logger.removeHandler(handler)
        handler.close()
    transcripts_formatter = logging.Formatter("%(message)s")
    transcripts_handler = logging.FileHandler(logs_path / f"transcript_{timestamp}.log")
    transcripts_handler.setFormatter(transcripts_formatter)
    transcripts_logger.addHandler(transcripts_handler)
    transcripts_logger.setLevel(logging.INFO)
    transcripts_logger.propagate = False



import argparse


def common_benchmark_parser():
    parser = argparse.ArgumentParser(description="Run exercism benchmarks")
    parser.add_argument(
        "--refresh_repo",
        action="store_true",
        default=False,
        help="When set local changes will be discarded.",
    )
    parser.add_argument(
        "--language",
        default="python",
        type=str,
        help="Which exercism language to do exercises for",
    )
    parser.add_argument(
        "--benchmarks",
        nargs="*",
        default=[],
        help=("Which benchmarks to run. max_benchmarks ignored when set. Exact meaning" " depends on benchmark."),
    )
    parser.add_argument(
        "--directory",
        "-d",
        default="benchmarks/benchmarks",
        type=str,
        help="Directory from which to load benchmarks and/or samples",
    )
    parser.add_argument(
        "--max_benchmarks",
        default=1,
        type=int,
        help="The maximum number of exercises to run",
    )
    parser.add_argument(
        "--max_iterations",
        default=1,
        type=int,
        help="Number of times to rerun mentat with error messages",
    )
    parser.add_argument(
        "--max_workers",
        default=1,
        type=int,
        help="Number of workers to use for multiprocessing",
    )
    parser.add_argument(
        "--retries",
        action="store",
        default=1,
        type=int,
        help="Number of times to retry a benchmark",
    )
    parser.add_argument(
        "--repo",
        action="store",
        default="mentat",
        help="For benchmarks that are evaluated against a repo",
    )
    parser.add_argument(
        "--evaluate_baseline",
        action="store_true",
        help="Evaluate the baseline for the benchmark",
    )
    parser.add_argument(
        "--swe_bench",
        nargs="?",
        const="dev",
        default=None,
        type=str,
        help="Fetch or load SWE-bench examples from split: dev (default), train or test.",
    )

    return parser



from typing import List

from typing_extensions import override

from mentat.auto_completer import get_command_filename_completions
from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path


class IncludeCommand(Command, command_name="include"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context

        if len(args) == 0:
            stream.send("No files specified", style="warning")
            return
        for file_path in args:
            included_paths = code_context.include(file_path)
            for included_path in included_paths:
                rel_path = get_relative_path(included_path, session_context.cwd)
                stream.send(f"{rel_path} added to context", style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("required", ["path", "glob pattern"], repeatable=True)]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return get_command_filename_completions(arguments[-1])

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Add files to context."



from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.git_handler import commit


class CommitCommand(Command, command_name="commit"):
    default_message = "Automatic commit"

    @override
    async def apply(self, *args: str) -> None:
        if args:
            commit(args[0])
        else:
            commit(self.__class__.default_message)

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("optional", "commit message")]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Commit all unstaged and staged changes to git."



from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Event
from pathlib import Path
from typing import AsyncIterator

import attr
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.completion_create_params import ResponseFormat

from mentat.code_file_manager import CodeFileManager
from mentat.errors import ModelError
from mentat.llm_api_handler import chunk_to_lines
from mentat.parsers.change_display_helper import (
    DisplayInformation,
    FileActionType,
    get_file_name_display,
    get_later_lines,
    get_previous_lines,
    get_removed_lines,
)
from mentat.parsers.file_edit import FileEdit
from mentat.parsers.streaming_printer import FormattedString, StreamingPrinter
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import convert_string_to_asynciter


@attr.define
class ParsedLLMResponse:
    full_response: str = attr.field()
    conversation: str = attr.field()
    file_edits: list[FileEdit] = attr.field()
    interrupted: bool = attr.field(default=False)


class Parser(ABC):
    def __init__(self):
        self.shutdown = Event()
        self._silence_printer = False

    @abstractmethod
    def get_system_prompt(self) -> str:
        pass

    def response_format(self) -> ResponseFormat:
        return ResponseFormat(type="text")

    async def stream_and_parse_llm_response(self, response: AsyncIterator[ChatCompletionChunk]) -> ParsedLLMResponse:
        """
        This general parsing structure relies on the assumption that all formats require three types of lines:
        1. 'conversation' lines, which are streamed as they come,
        2. 'special' lines, that are never shown to the user and contain information such as the file_name
        3. 'code' lines, which are the actual code written and are shown to the user in a special format
        To make a parser that differs from these assumptions, override this method instead of the helper methods
        """
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_file_manager = session_context.code_file_manager

        printer = StreamingPrinter()
        if self._silence_printer:
            printer_task = None
        else:
            printer_task = asyncio.create_task(printer.print_lines())
        message = ""
        conversation = ""
        file_edits = dict[Path, FileEdit]()

        cur_line = ""
        prev_block = ""
        cur_block = ""
        display_information: DisplayInformation | None = None
        file_edit: FileEdit | None = None
        line_printed = False
        in_special_lines = False
        in_code_lines = False
        in_conversation = True
        rename_map = dict[Path, Path]()
        interrupted = False
        async for chunk in response:
            if self.shutdown.is_set():
                interrupted = True
                printer.shutdown_printer()
                if printer_task is not None:
                    await printer_task
                stream.send("\nInterrupted by user. Using the response up to this point.")
                break

            for content in chunk_to_lines(chunk):
                if not content:
                    continue
                message += content
                cur_line += content

                # Print if not in special lines and line is confirmed not special
                if not in_special_lines:
                    if not line_printed:
                        if not self._could_be_special(cur_line):
                            line_printed = True
                            if not in_code_lines or display_information is None:
                                printer.add_string(cur_line, end="")
                                conversation += cur_line
                            else:
                                printer.add_string(
                                    self._code_line_beginning(display_information, cur_block),
                                    end="",
                                )
                                printer.add_string(
                                    self._code_line_content(
                                        display_information,
                                        cur_line,
                                        cur_line,
                                        cur_block,
                                    ),
                                    end="",
                                )
                    else:
                        if not in_code_lines or display_information is None:
                            printer.add_string(content, end="")
                            conversation += content
                        else:
                            printer.add_string(
                                self._code_line_content(display_information, content, cur_line, cur_block),
                                end="",
                            )

                # If we print non code lines, we want to reprint the file name of the next change,
                # even if it's the same file as the last change
                if not in_code_lines and not in_special_lines and line_printed:
                    printer.cur_file = None
                    printer.cur_file_display = None
                    file_edit = None
                    display_information = None
                    in_conversation = True

                # New line handling
                if "\n" in cur_line:
                    # Now that full line is in, give _could_be_special full line (including newline)
                    # and see if it should be printed or not
                    if not in_special_lines and not line_printed and not self._could_be_special(cur_line):
                        if not in_code_lines or display_information is None:
                            printer.add_string(cur_line, end="")
                        else:
                            printer.add_string(
                                self._code_line_beginning(display_information, cur_block),
                                end="",
                            )
                            printer.add_string(
                                self._code_line_content(display_information, cur_line, cur_line, cur_block),
                                end="",
                            )
                        line_printed = True

                    if self._starts_special(cur_line.strip()):
                        in_special_lines = True

                    if in_special_lines or in_code_lines:
                        cur_block += cur_line

                    if in_special_lines and self._ends_special(cur_line.strip()):
                        previous_file = None if file_edit is None else file_edit.file_path
                        previous_file_had_edits = (
                            False
                            if file_edit is None
                            else file_edit.replacements or file_edit.is_creation or file_edit.is_deletion
                        )

                        try:
                            (
                                display_information,
                                file_edit,
                                in_code_lines,
                            ) = self._special_block(
                                code_file_manager,
                                session_context.cwd,
                                rename_map,
                                cur_block,
                            )
                        except ModelError as e:
                            printer.add_string((str(e), {"color": "red"}))
                            printer.add_string("Using existing changes.")
                            printer.wrap_it_up()
                            if printer_task is not None:
                                await printer_task
                            logging.debug("LLM Response:")
                            logging.debug(message)
                            return ParsedLLMResponse(
                                message,
                                conversation,
                                [file_edit for file_edit in file_edits.values()],
                            )

                        # Rename map handling
                        if file_edit.rename_file_path is not None:
                            rename_map[file_edit.rename_file_path] = file_edit.file_path
                        if file_edit.file_path in rename_map:
                            file_edit.file_path = session_context.cwd / rename_map[file_edit.file_path]

                        # Add a delimiter directly before a new file edit if it's the same file as before
                        # This way, we get delimiters between every edit but not before or after the whole thing.
                        if previous_file == file_edit.file_path and previous_file_had_edits:
                            printer.add_delimiter()

                        printer.cur_file = str(file_edit.file_path)
                        printer.cur_file_display = get_file_name_display(display_information)
                        in_special_lines = False
                        prev_block = cur_block
                        cur_block = ""

                        # New file_edit creation and merging
                        if file_edit.file_path not in file_edits:
                            file_edits[file_edit.file_path] = file_edit
                        else:
                            cur_file_edit = file_edits[file_edit.file_path]
                            cur_file_edit.is_creation = cur_file_edit.is_creation or file_edit.is_creation
                            cur_file_edit.is_deletion = cur_file_edit.is_deletion or file_edit.is_deletion
                            if file_edit.rename_file_path is not None:
                                cur_file_edit.rename_file_path = file_edit.rename_file_path
                            cur_file_edit.replacements.extend(file_edit.replacements)
                            file_edit = cur_file_edit

                        # Send empty string to start filename block; needed in case it's a rename,
                        # in which case this is all that will be sent from this fileedit)
                        if (
                            in_conversation
                            or display_information.file_action_type == FileActionType.RenameFile
                            or (file_edit.file_path != previous_file)
                        ):
                            in_conversation = False
                            printer.add_string("", end="", allow_empty=True)

                        # Print previous lines, removed block, and possibly later lines
                        if in_code_lines or display_information.removed_block:
                            printer.add_string(get_previous_lines(display_information))
                            printer.add_string(get_removed_lines(display_information))
                            if not in_code_lines:
                                printer.add_string(get_later_lines(display_information))
                    elif in_code_lines and self._ends_code(cur_line.strip()):
                        # Adding code lines to previous file_edit and printing later lines
                        if display_information is not None and file_edit is not None:
                            self._add_code_block(
                                code_file_manager,
                                rename_map,
                                prev_block,
                                cur_block,
                                display_information,
                                file_edit,
                            )
                            printer.add_string(get_later_lines(display_information))

                        in_code_lines = False
                        prev_block = cur_block
                        cur_block = ""
                    line_printed = False
                    cur_line = ""
        else:
            # If the model doesn't close out the code lines, we might as well do it for it
            if in_code_lines and display_information is not None and file_edit is not None:
                self._add_code_block(
                    code_file_manager,
                    rename_map,
                    prev_block,
                    cur_block,
                    display_information,
                    file_edit,
                )
                printer.add_string(get_later_lines(display_information))

            # Only finish printing if we don't quit from ctrl-c
            printer.wrap_it_up()
            if printer_task is not None:
                await printer_task

        logging.debug("LLM Response:")
        logging.debug(message)
        return ParsedLLMResponse(
            message,
            conversation,
            [file_edit for file_edit in file_edits.values()],
            interrupted,
        )

    # Ideally this would be called in this class instead of subclasses
    def _get_file_lines(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        abs_path: Path,
    ) -> list[str]:
        path = rename_map.get(
            abs_path,
            abs_path,
        )
        return code_file_manager.file_lines.get(path, []).copy()

    # These methods aren't abstract, since most parsers will use this implementation, but can be overriden easily
    def provide_line_numbers(self) -> bool:
        return True

    def line_number_starting_index(self) -> int:
        return 1

    def _code_line_beginning(self, display_information: DisplayInformation, cur_block: str) -> FormattedString:
        """
        The beginning of a code line; normally this means printing the + prefix
        """
        return (
            "+" + " " * (display_information.line_number_buffer - 1),
            {"color": "green"},
        )

    def _code_line_content(
        self,
        display_information: DisplayInformation,
        content: str,
        cur_line: str,
        cur_block: str,
    ) -> FormattedString:
        """
        Part of a code line; normally this means printing in green
        """
        return (content, {"color": "green"})

    # These methods must be overriden if using the default stream and parse function
    def _could_be_special(self, cur_line: str) -> bool:
        """
        Returns if this current line could be a special line and therefore shouldn't be printed yet.
        Once line is completed, will include a newline character.
        """
        raise NotImplementedError()

    def _starts_special(self, line: str) -> bool:
        """
        Determines if this line begins a special block
        """
        raise NotImplementedError()

    def _ends_special(self, line: str) -> bool:
        """
        Determines if this line ends a special block
        """
        raise NotImplementedError()

    def _special_block(
        self,
        code_file_manager: CodeFileManager,
        cwd: Path,
        rename_map: dict[Path, Path],
        special_block: str,
    ) -> tuple[DisplayInformation, FileEdit, bool]:
        """
        After finishing special block, return DisplayInformation to print, FileEdit to add/merge to list,
        and if a code block follows this special block.
        """
        raise NotImplementedError()

    def _ends_code(self, line: str) -> bool:
        """
        Determines if this line ends a code block
        """
        raise NotImplementedError()

    def _add_code_block(
        self,
        code_file_manager: CodeFileManager,
        rename_map: dict[Path, Path],
        special_block: str,
        code_block: str,
        display_information: DisplayInformation,
        file_edit: FileEdit,
    ) -> None:
        """
        Using the special block, code block and display_information, edits the FileEdit to add the new code block.
        """
        raise NotImplementedError()

    async def parse_llm_response(self, response: str) -> ParsedLLMResponse:
        self._silence_printer = True
        async_iter_response = convert_string_to_asynciter(response, chunk_size=100)
        parsed_response = await self.stream_and_parse_llm_response(async_iter_response)
        self._silence_printer = False
        return parsed_response



from enum import Enum
from pathlib import Path
from typing import List, Tuple, cast

import attr
from pygments import lex
from pygments.formatters import TerminalFormatter
from pygments.lexer import Lexer
from pygments.lexers import TextLexer, get_lexer_for_filename
from pygments.util import ClassNotFound

from mentat.parsers.streaming_printer import FormattedString
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path


def get_lexer(file_path: Path):
    try:
        lexer: Lexer = get_lexer_for_filename(file_path)
    except ClassNotFound:
        lexer = TextLexer()
    lexer.stripnl = False
    lexer.stripall = False
    lexer.ensurenl = False
    return lexer


def get_line_number_buffer(file_lines: list[str]):
    return 1


class FileActionType(Enum):
    RenameFile = "rename"
    CreateFile = "create"
    DeleteFile = "delete"
    UpdateFile = "update"


def get_file_action_type(is_creation: bool, is_deletion: bool, new_name: Path | None):
    if is_creation:
        file_action_type = FileActionType.CreateFile
    elif is_deletion:
        file_action_type = FileActionType.DeleteFile
    elif new_name is not None:
        file_action_type = FileActionType.RenameFile
    else:
        file_action_type = FileActionType.UpdateFile
    return file_action_type


@attr.define(slots=False)
class DisplayInformation:
    file_name: Path = attr.field()
    file_lines: list[str] = attr.field()
    added_block: list[str] = attr.field()
    removed_block: list[str] = attr.field()
    file_action_type: FileActionType = attr.field()
    first_changed_line: int = attr.field(default=0)
    last_changed_line: int = attr.field(default=0)
    new_name: Path | None = attr.field(default=None)

    def __attrs_post_init__(self):
        ctx = SESSION_CONTEXT.get()

        self.line_number_buffer = get_line_number_buffer(self.file_lines)
        self.lexer = get_lexer(self.file_name)

        if self.file_name.is_absolute():
            self.file_name = get_relative_path(self.file_name, ctx.cwd)
        if self.new_name is not None and self.new_name.is_absolute():
            self.new_name = get_relative_path(self.new_name, ctx.cwd)


def _remove_empty_lines(lines: list[str]) -> list[str]:
    if not lines:
        return []

    # Find the first non-empty line
    start = 0
    while start < len(lines) and not lines[start].strip():
        start += 1

    # Find the last non-empty line
    end = len(lines) - 1
    while end > start and not lines[end].strip():
        end -= 1

    # If all lines are empty, return nothing
    if start == len(lines):
        return []

    # Return the list with no empty lines
    return lines[start : end + 1]


def _prefixed_lines(line_number_buffer: int, lines: list[str], prefix: str) -> str:
    return "\n".join([prefix + " " * (line_number_buffer - len(prefix)) + line.strip("\n") for line in lines])


def _get_code_block(
    code_lines: list[str],
    line_number_buffer: int,
    prefix: str,
    color: str | None,
) -> FormattedString:
    lines = _prefixed_lines(line_number_buffer, code_lines, prefix)
    if color is None:
        return lines
    else:
        return (lines, {"color": color})


def display_full_change(display_information: DisplayInformation, prefix: str = ""):
    ctx = SESSION_CONTEXT.get()

    full_change: List[FormattedString] = [
        get_file_name(display_information),
        ("=" * 60 if display_information.added_block or display_information.removed_block else ""),
        get_previous_lines(display_information),
        get_removed_lines(display_information),
        get_added_lines(display_information),
        get_later_lines(display_information),
        ("=" * 60 if display_information.added_block or display_information.removed_block else ""),
    ]
    # TODO: This is called by undo and redo; this is very hardcoded to terminalclient ui right now
    # (so it might look weird in vscode extension), but once we move commands to terminalclient we can remove this
    for line in full_change:
        if isinstance(line, str):
            if not line.strip():
                continue
            ctx.stream.send(prefix, end="")
            ctx.stream.send(f"\n{prefix}".join(line.split("\n")))
        elif isinstance(line, Tuple):
            if not line[0].strip():
                continue
            for sub_line in line[0].split("\n"):
                ctx.stream.send(prefix, end="")
                ctx.stream.send(sub_line, **line[1])
        else:
            ctx.stream.send(prefix, end="")
            for text in line:
                for i, sub_line in enumerate(text[0].split("\n")):
                    if i != 0:
                        ctx.stream.send(prefix, end="")
                    ctx.stream.send(sub_line, **text[1], end="")
                    if i != len(text[0].split("\n")) - 1:
                        ctx.stream.send("")
            ctx.stream.send("")


def get_file_name(
    display_information: DisplayInformation,
) -> FormattedString:
    match display_information.file_action_type:
        case FileActionType.CreateFile:
            return (f"\n{display_information.file_name}*", {"color": "bright_green"})
        case FileActionType.DeleteFile:
            return (
                f"\nDeletion: {display_information.file_name}",
                {"color": "bright_red"},
            )
        case FileActionType.RenameFile:
            return (
                (f"\nRename: {display_information.file_name} ->" f" {display_information.new_name}"),
                {"color": "yellow"},
            )
        case FileActionType.UpdateFile:
            return (f"\n{display_information.file_name}", {"color": "bright_blue"})


def get_file_name_display(display_information: DisplayInformation):
    match display_information.file_action_type:
        case FileActionType.CreateFile:
            return (f"{display_information.file_name}*", "creation")
        case FileActionType.DeleteFile:
            return (
                f"Deletion: {display_information.file_name}",
                "deletion",
            )
        case FileActionType.RenameFile:
            return (
                (f"Rename: {display_information.file_name} ->" f" {display_information.new_name}"),
                "rename",
            )
        case FileActionType.UpdateFile:
            return (f"{display_information.file_name}", "edit")


def get_added_lines(
    display_information: DisplayInformation,
    prefix: str = "+",
    color: str | None = "green",
) -> FormattedString:
    return _get_code_block(
        display_information.added_block,
        display_information.line_number_buffer,
        prefix,
        color,
    )


def get_removed_lines(
    display_information: DisplayInformation,
    prefix: str = "-",
    color: str | None = "red",
) -> FormattedString:
    return _get_code_block(
        display_information.removed_block,
        display_information.line_number_buffer,
        prefix,
        color,
    )


def highlight_text(text: str, lexer: Lexer) -> FormattedString:
    formatter = TerminalFormatter(bg="dark")  # type: ignore
    string: FormattedString = []
    for ttype, value in lex(text, lexer):
        # We use TerminalFormatter's color scheme; TODO: Hook this up to our style themes instead
        color = cast(str, formatter._get_color(ttype))  # type: ignore

        # Convert Pygment styles to Rich styles
        if color.startswith("*"):
            # TODO: Send bold style
            color = color[1:-1]
        if color.startswith("_"):
            # TODO: Send italic style
            color = color[1:-1]
        if color.startswith("bright"):
            color = color.replace("bright", "bright_")

        if not color:
            color = None

        string.append((value, {"color": color}))
    return string


def get_previous_lines(
    display_information: DisplayInformation,
    num: int = 2,
) -> FormattedString:
    if display_information.first_changed_line < 0:
        return ""
    lines = _remove_empty_lines(
        [
            display_information.file_lines[i]
            for i in range(
                max(0, display_information.first_changed_line - num),
                min(
                    display_information.first_changed_line,
                    len(display_information.file_lines),
                ),
            )
        ]
    )
    buffered = [" " * display_information.line_number_buffer + line for line in lines]

    prev = "\n".join(buffered)
    return highlight_text(prev, display_information.lexer)


def get_later_lines(
    display_information: DisplayInformation,
    num: int = 2,
) -> FormattedString:
    if display_information.last_changed_line < 0:
        return ""
    lines = _remove_empty_lines(
        [
            display_information.file_lines[i]
            for i in range(
                max(0, display_information.last_changed_line),
                min(
                    display_information.last_changed_line + num,
                    len(display_information.file_lines),
                ),
            )
        ]
    )
    buffered = [" " * (display_information.line_number_buffer) + line for line in lines]

    later = "\n".join(buffered)
    return highlight_text(later, display_information.lexer)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Literal

import attr

from mentat.errors import MentatError
from mentat.session_context import SESSION_CONTEXT


@attr.define
class CommandArgument:
    # Required: <arg>, Optional: [arg], Literal: arg
    arg_type: Literal["required", "optional", "literal"] = attr.field()
    # Multiple descriptions will be separated with |: desc1|desc2
    description: str | List[str] = attr.field()
    # Repeatable arguments will be followed with ...: <arg1> ...
    repeatable: bool = attr.field(default=False)


class Command(ABC):
    """
    Base Command class. To create a new command, extend this class, provide a command_name,
    and import the class in commands.__init__.py so that it is initialized on startup.
    """

    # Unfortunately, Command isn't defined here yet, so even with annotations we need quotation marks
    _registered_commands = dict[str, type["Command"]]()
    hidden = False

    def __init_subclass__(cls, command_name: str | None) -> None:
        if command_name is not None:
            Command._registered_commands[command_name] = cls

    @classmethod
    def create_command(cls, command_name: str) -> Command:
        if command_name not in cls._registered_commands:
            return InvalidCommand(command_name)

        command_cls = cls._registered_commands[command_name]
        return command_cls()

    @classmethod
    def get_command_names(cls) -> list[str]:
        return [name for name, command in cls._registered_commands.items() if not command.hidden]

    @abstractmethod
    async def apply(self, *args: str) -> None:
        pass

    @classmethod
    @abstractmethod
    def arguments(cls) -> List[CommandArgument]:
        pass

    @classmethod
    @abstractmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        """
        Returns a list of possible completions for the argument in the specified position (0-indexed)
        given the previous arguments
        """
        pass

    @classmethod
    @abstractmethod
    def help_message(cls) -> str:
        pass


class InvalidCommand(Command, command_name=None):
    def __init__(self, invalid_name: str):
        self.invalid_name = invalid_name

    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream

        stream.send(
            f"{self.invalid_name} is not a valid command. Use /help to see a list of" " all valid commands",
            style="warning",
        )

    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        raise MentatError("Argument names called on invalid command")

    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        raise MentatError("Help message called on invalid command")



from __future__ import annotations

import asyncio
import logging
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Dict, List, Literal, cast
from uuid import UUID, uuid4

from pydantic import BaseModel

from mentat.broadcast import Broadcast


class StreamMessageSource:
    # Enums can't be serialized or deserialized, since Enum.value is an instance of Enum, not the actual value
    TYPE = Literal["server", "client"]
    SERVER = "server"
    CLIENT = "client"


class StreamMessage(BaseModel):
    id: UUID
    channel: str
    source: StreamMessageSource.TYPE
    data: Any
    extra: Dict[str, Any]


class SessionStream:
    """
    Used to send and receive messages from the client.

    Channels and their expected behavior (starred channels are sent by the client):

    default: Any data sent to the client over this channel should be displayed.
    Valid kwargs: color, style, filepath, filepath_display, delimiter

    *session_exit: Sent by the client, suggesting that the session should exit whenever possible.
    client_exit: Sent by the server, client should shut down when received.
    session_stopped: Sent by the server directly before server shuts down. Server can't be contacted after received.

    loading: Used to tell the client to display a loading bar. Valid kwargs: terminate

    input_request: Used to request input from the client (data unused). Valid kwargs: command_autocomplete
    *input_request:<message_id>: Sent by the client. The channel the response to an input_request is sent over.

    model_file_edits: Once the model fully completes its response, sends a list of the file edits. Schema:
    [
        {
            "file_path": "/absolute/path/to/file",
            "new_file_path": "/absolute/path/to/renamed_file", # May be null
            "type": "edit" | "creation" | "deletion",
            "new_content": "The new file text after edits are applied"
        },
        ...
    ]

    edits_complete: A boolean sent when edits have been completed. True if any edits were accepted.

    *completion_request: Sent by the client, retrieves completions for given data. Valid kwargs: command_autocomplete
    completion_request:<message_id>: The response to the given completion request.

    default_prompt: The prefilled prompt to show on next user input request. Should be additive and reset
    after every input request. See TerminalClient for exact implementation.

    interruptable: A boolean sent to enable or disable an 'interrupt' button.
    If an interrupt is sent while interruptable is false, the server will shut down.
    *interrupt: Sent by the client. Sent whenever client interrupts current work. Sent by things like Ctrl-C.

    context_update: An object describing the context sent whenever the context changes. Schema:
    {
        "cwd": "Mentat's cwd",
        "diff_context_display": "The display for the diff context",
        "auto_context_tokens": The number of auto context tokens,
        "features": ["List of user included features"],
        "git_diff_paths": ["List of all paths with git diffs; used to color the changed included features"],
        "total_tokens": Total tokens in context,
        "maximum_tokens": Maximum tokens allowed in context,
        "total_cost": Total cost so far
    }

    *include: Sent by the client. Gives a path to include in context.
    *exclude: Sent by the client. Gives a path to exclude from context.
    *clear_conversation: Sent by the client. Clears the conversation history (but not the context).
    """

    def __init__(self):
        self.messages: List[StreamMessage] = []
        self._interrupt_lock = asyncio.Lock()
        self._broadcast = Broadcast()

    def start(self):
        self._broadcast.connect()

    def stop(self):
        self._broadcast.disconnect()

    # Since there is no maximum queue size, use the synchronous version of this function instead
    async def send_async(
        self,
        data: Any,
        source: StreamMessageSource.TYPE = StreamMessageSource.SERVER,
        channel: str = "default",
        **kwargs: Any,
    ):
        message = StreamMessage(
            id=uuid4(),
            source=source,
            channel=channel,
            data=data,
            extra=kwargs,
        )

        self.messages.append(message)
        await self._broadcast.publish_async(channel=channel, message=message)

        return message

    def send(
        self,
        data: Any,
        source: StreamMessageSource.TYPE = StreamMessageSource.SERVER,
        channel: str = "default",
        **kwargs: Any,
    ):
        message = StreamMessage(
            id=uuid4(),
            source=source,
            channel=channel,
            data=data,
            extra=kwargs,
        )

        self.messages.append(message)
        self._broadcast.publish(channel=channel, message=message)

        return message

    def send_stream_message(self, message: StreamMessage):
        self.messages.append(message)
        self._broadcast.publish(channel=message.channel, message=message)

    async def recv(self, channel: str = "default") -> StreamMessage:
        """Listen for a single event on a channel"""
        with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                stream_message = cast(StreamMessage, event.message)
                return stream_message
            raise Exception("recv should not complete without receiving an Event")

    async def listen(self, channel: str = "default") -> AsyncGenerator[StreamMessage, None]:
        """Listen to all messages on a channel indefinitely"""
        with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                yield event.message

    async def universal_listen(self) -> AsyncGenerator[StreamMessage, None]:
        with self._broadcast.universal_subscribe() as subscriber:
            async for event in subscriber:
                yield event.message

    async def join(self) -> None:
        """Blocks until all sent events have been processed"""
        await self._broadcast.join()

    @asynccontextmanager
    async def interrupt_catcher(self, event: asyncio.Event):
        """
        Will start a task listening for an interrupt, set the given event when one is received, and clear it at the end
        """

        interrupt_task = asyncio.create_task(self._listen_for_interrupt(event))
        self.send(True, channel="interruptable")
        yield
        self.send(False, channel="interruptable")
        interrupt_task.cancel()
        try:
            await interrupt_task
        except asyncio.CancelledError:
            pass
        event.clear()

    async def _listen_for_interrupt(self, event: asyncio.Event):
        """
        Listens for an interrupt message from the client. Set the event given when an intterupt is received.
        """

        async with self._interrupt_lock:
            await self.recv("interrupt")
            logging.info("User sent interrupt.")
            event.set()

    def is_interrupt_locked(self) -> bool:
        return self._interrupt_lock.locked()



from pathlib import Path
from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.errors import SampleError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path


class SampleCommand(Command, command_name="sample"):
    hidden = True  # Waiting for parse_message to be complete

    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        sampler = session_context.sampler
        stream = session_context.stream
        try:
            sample = await sampler.create_sample()
        except SampleError as e:
            stream.send(f"Failed to generate sample: {e}", style="error")
            return
        fname = f"sample_{sample.id}.json"
        if len(args) > 0:
            fpath = Path(args[0]) / fname
        else:
            samples_dir = mentat_dir_path / "samples"
            samples_dir.mkdir(exist_ok=True)
            fpath = samples_dir / fname
        sample.save(fpath)
        SESSION_CONTEXT.get().stream.send(f"Sample saved to {fpath}.", style="success")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [CommandArgument("optional", "path")]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Save a sample of the current session."



from pathlib import Path

from benchmarks.exercise_runners.abstract_exercise_runner import AbstractExerciseRunner


class ClojureExerciseRunner(AbstractExerciseRunner):
    def __init__(self, exercise):
        super().__init__(exercise, "clj")
        self.file = Path(exercise.replace("-", "_") + ".clj")
        self.full_path = self.dir / "src" / self.file

    def run_test(self):
        self._run_test_command(["lein", "test", ":only", self.name + "-test"], cwd=self.dir)

    def passed(self):
        try:
            with open(self.test_output_file, "r") as f:
                lines = f.readlines()
                return "0 failures, 0 errors." in lines[-1]
        except FileNotFoundError:
            return False



from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class AmendCommand(Command, command_name="amend"):
    @override
    async def apply(self, *args: str) -> None:
        ctx = SESSION_CONTEXT.get()

        last_message = ctx.conversation.amend()
        if last_message:
            ctx.stream.send("Previous response removed from context.", style="info")
            ctx.stream.send(last_message, channel="default_prompt")
        else:
            ctx.stream.send("No previous user requests to amend.", style="warning")

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return (
            "Used to amend a previous user request. Works by resetting context to the"
            " state it was at the last request and prefills user input with the last"
            " request. Does not undo any edits."
        )



from typing import List

from typing_extensions import override

from mentat.auto_completer import get_command_filename_completions
from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class RunCommand(Command, command_name="run"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        conversation = session_context.conversation
        if not args:
            stream.send("No command given", style="failure")
            return
        await conversation.run_command(list(args))

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return [
            CommandArgument("required", "command"),
            CommandArgument("optional", "args", repeatable=True),
        ]

    @override
    @classmethod
    def argument_autocompletions(cls, arguments: list[str], argument_position: int) -> list[str]:
        return get_command_filename_completions(arguments[-1])

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Run a shell command and put its output in context."



from timeit import default_timer

import chromadb
from chromadb.api.types import Embeddable, EmbeddingFunction, Embeddings

from mentat.code_feature import CodeFeature, count_feature_tokens
from mentat.errors import MentatError
from mentat.llm_api_handler import model_context_size, model_price_per_1000_tokens
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no
from mentat.utils import mentat_dir_path

EMBEDDINGS_API_BATCH_SIZE = 2048

client = chromadb.PersistentClient(path=str(mentat_dir_path / "chroma"))


class MentatEmbeddingFunction(EmbeddingFunction[Embeddable]):
    def __call__(self, input: Embeddable) -> Embeddings:
        if not all(isinstance(item, str) for item in input):
            raise MentatError("MentatEmbeddings only enabled for text files")
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        llm_api_handler = session_context.llm_api_handler

        n_batches = 0 if len(input) == 0 else len(input) // EMBEDDINGS_API_BATCH_SIZE + 1
        output: Embeddings = []
        for batch in range(n_batches):
            i_start, i_end = (
                batch * EMBEDDINGS_API_BATCH_SIZE,
                (batch + 1) * EMBEDDINGS_API_BATCH_SIZE,
            )
            response = llm_api_handler.call_embedding_api(input[i_start:i_end], config.embedding_model)
            output += response
        return output


class Collection:
    def __init__(self, embedding_model: str):
        self._collection = client.get_or_create_collection(
            name=f"mentat-{embedding_model}",
            embedding_function=MentatEmbeddingFunction(),
        )

    def exists(self, id: str) -> bool:
        assert self._collection is not None, "Collection not initialized"
        return len(self._collection.get(id)["ids"]) > 0

    def add(self, checksums: list[str], texts: list[str]) -> None:
        assert self._collection is not None, "Collection not initialized"
        return self._collection.add(  # type: ignore
            ids=checksums,
            documents=texts,
            metadatas=[{"active": False} for _ in checksums],
        )

    def query(self, prompt: str, checksums: list[str]) -> dict[str, float]:
        assert self._collection is not None, "Collection not initialized"

        self._collection.update(  # type: ignore
            ids=checksums,
            metadatas=[{"active": True} for _ in checksums],
        )
        results = self._collection.query(  # type: ignore
            query_texts=[prompt],
            where={"active": True},
            n_results=len(checksums) + 1,
        )
        self._collection.update(  # type: ignore
            ids=checksums,
            metadatas=[{"active": False} for _ in checksums],
        )
        assert results["distances"], "Error calculating distances"
        return {c: e for c, e in zip(results["ids"][0], results["distances"][0])}


async def get_feature_similarity_scores(
    prompt: str,
    features: list[CodeFeature],
) -> list[float]:
    """Return the similarity scores for a given prompt and list of features."""
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    config = session_context.config
    cost_tracker = session_context.cost_tracker
    embedding_model = session_context.config.embedding_model

    max_model_tokens = model_context_size(config.embedding_model)
    if max_model_tokens is None:
        raise MentatError(f"Missing model context size for {embedding_model}.")

    # Initialize DB
    collection = Collection(embedding_model)

    # Identify which items need embeddings.
    checksums: list[str] = [f.get_checksum() for f in features]
    tokens: list[int] = await count_feature_tokens(features, embedding_model)
    embed_texts = list[str]()
    embed_checksums = list[str]()
    embed_tokens = list[int]()
    for feature, checksum, token in zip(features, checksums, tokens):
        if token > max_model_tokens:
            stream.send(
                f"Warning: Feature {str(feature)} has {token} tokens, which exceeds the"
                f" maximum of {max_model_tokens} for model {config.embedding_model}."
                " Skipping."
            )
            continue
        if not collection.exists(checksum) and checksum not in embed_checksums:
            embed_texts.append("\n".join(feature.get_code_message()))
            embed_checksums.append(checksum)
            embed_tokens.append(token)

    # If it costs more than $1, get confirmation from user.
    cost = model_price_per_1000_tokens(embedding_model)
    if cost is None:
        stream.send(
            "Warning: Could not determine cost of embeddings. Continuing anyway.",
            style="warning",
        )
    else:
        expected_cost = (sum(embed_tokens) / 1000) * cost[0]
        if expected_cost > 1.0:
            stream.send(f"Embedding {sum(embed_tokens)} tokens will cost ${cost[0]:.2f}." " Continue anyway?")
            if not await ask_yes_no(default_yes=True):
                stream.send("Ignoring embeddings for now.")
                return [0.0 for _ in checksums]

    # Load embeddings
    if embed_texts:
        start_time = default_timer()
        stream.send(None, channel="loading")
        collection.add(embed_checksums, embed_texts)
        cost_tracker.log_api_call_stats(
            sum(embed_tokens),
            0,
            embedding_model,
            start_time - default_timer(),
        )

    # Get similarity scores
    stream.send(None, channel="loading")
    _checksums = list(set(checksums))
    scores = collection.query(prompt, _checksums)
    return [scores.get(f.get_checksum(), 0) for f in features]



from mentat.code_feature import CodeFeature
from mentat.feature_filters.feature_filter import FeatureFilter


class EmbeddingSimilarityFilter(FeatureFilter):
    def __init__(self, query: str):
        self.query = query

    async def score(
        self,
        features: list[CodeFeature],
    ) -> list[tuple[CodeFeature, float]]:
        from mentat.embeddings import (  # dynamic import to improve startup time.
            get_feature_similarity_scores,
        )

        if self.query == "":
            return [(f, 0.0) for f in features]

        sim_scores = await get_feature_similarity_scores(self.query, features)
        features_scored = zip(features, sim_scores)
        return sorted(features_scored, key=lambda x: x[1])

    async def filter(
        self,
        features: list[CodeFeature],
    ) -> list[CodeFeature]:
        if self.query == "":
            return features
        return [f for f, _ in await self.score(features)]



from pathlib import Path

from git import Optional
from textual.suggester import Suggester
from typing_extensions import override


class HistorySuggester(Suggester):
    def __init__(self, history_file: Path, *, case_sensitive: bool = True) -> None:
        self.history_file = history_file
        self.position = -1
        if not history_file.exists():
            with open(self.history_file, "w") as f:
                pass
        with open(self.history_file, "r") as f:
            self._suggestions = f.read().split("\n")
        super().__init__(use_cache=False, case_sensitive=case_sensitive)

    def append_to_history(self, submission: str):
        self.position = 0
        if not submission.strip() or (self._suggestions and self._suggestions[-1] == submission):
            return
        self._suggestions.append(submission)
        with open(self.history_file, "a") as f:
            f.write(f"{submission}\n")

    @override
    async def get_suggestion(self, value: str) -> str | None:
        for suggestion in reversed(self._suggestions):
            if not self.case_sensitive:
                suggestion = suggestion.casefold()
                value = value.casefold()
            if suggestion.startswith(value):
                return suggestion
        return None

    def move_up(self) -> Optional[str]:
        if self.position == -len(self._suggestions):
            return None
        self.position -= 1
        return self._suggestions[self.position]

    def move_down(self) -> Optional[str]:
        if self.position == -1:
            return None
        self.position += 1
        return self._suggestions[self.position]

    # This is a bit of a hacky way to let the autocomplete (which watches the input value)
    # know if the new input value was from moving in history (in which case we don't want autocomplete to pop up)
    def just_moved(self, value: str):
        return self._suggestions[self.position] == value
